from gtfsdb import config
from gtfsdb import util
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from contextlib import contextmanager

import logging
log = logging.getLogger(__file__)


class Database(object):

    def __init__(self, **kwargs):
        """
        keyword arguments:
            is_geospatial: if database supports geo functions
            schema: database schema name
            tables: limited list of tables to load into database
            url: SQLAlchemy database url
        """
        self.tables = kwargs.get('tables', None)
        url = kwargs.get('url')
        if not url:
            url = kwargs.get('database_url', config.DEFAULT_DATABASE_URL)
        self.url = url
        self.schema = kwargs.get('schema', config.DEFAULT_SCHEMA)
        self.is_geospatial = kwargs.get('is_geospatial', config.DEFAULT_IS_GEOSPATIAL)

        """Order list of class names, used for creating & populating tables"""
        from gtfsdb import SORTED_CLASS_NAMES, CURRENT_CLASS_NAMES
        self.sorted_class_names = SORTED_CLASS_NAMES
        if kwargs.get('current_tables'):
            self.sorted_class_names.extend(CURRENT_CLASS_NAMES)
        # import pdb; pdb.set_trace()

    @property
    def classes(self):
        subclasses = self.get_base_subclasses()
        if self.tables:
            ret_val = [c for c in subclasses if c.__tablename__ in self.tables]
        else:
            ret_val = subclasses
        return ret_val

    @property
    def sorted_classes(self):
        classes = []
        for class_name in self.sorted_class_names:
            cls = next((c for c in self.classes if c.__name__ == class_name), None)
            if cls:
                classes.append(cls)
        return classes

    @classmethod
    def get_base_subclasses(cls):
        from gtfsdb.model.base import Base
        return util.get_all_subclasses(Base)

    @property
    def metadata(self):
        from gtfsdb.model.base import Base
        return Base.metadata

    def load_tables(self, **kwargs):
        """ load the sorted classes """
        for cls in self.sorted_classes:
            cls.load(self, **kwargs)

    def postprocess_tables(self, **kwargs):
        """ call the post-process routines on the sorted classes """
        do_postprocess = kwargs.get('do_postprocess', True)
        if do_postprocess:
            for cls in self.sorted_classes:
                cls.post_process(self, **kwargs)

    def create(self):
        """ drop/create GTFS database """
        for cls in self.sorted_classes:
            self.create_table(cls)

    def create_table(self, orm_class, check_first=True, drop_first=True):
        log.debug("create table: {0}".format(orm_class.__table__))
        try:
            if drop_first:
                orm_class.__table__.drop(self.engine, checkfirst=check_first)
        except:
            log.info("NOTE: couldn't *drop* table {0} (might not be a big deal)".format(orm_class.__table__))
        try:
            orm_class.__table__.create(self.engine, checkfirst=check_first)
        except Exception as e:
            log.info("NOTE: couldn't *create* table {0} (could be a big deal)\n{1}".format(orm_class.__table__, e))

    @property
    def dialect_name(self):
        return self.engine.url.get_dialect().name

    @property
    def is_geospatial(self):
        return self._is_geospatial

    @is_geospatial.setter
    def is_geospatial(self, val):
        self._is_geospatial = val
        for cls in self.classes:
            if val and hasattr(cls, 'add_geometry_column'):
                cls.add_geometry_column()

    @property
    def schema(self):
        return self._schema

    @schema.setter
    def schema(self, val):
        self._schema = val

        # TODO ... move to create() method
        try:
            if self._schema:
                from sqlalchemy.schema import CreateSchema
                self.engine.execute(CreateSchema(self._schema), checkfirst=True)
        except Exception as e:
            log.info("NOTE: couldn't create schema {0} (schema might already exist)\n{1}".format(self._schema, e))

        for cls in self.classes:
            cls.set_schema(self._schema)

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, val):
        self._url = val
        self.engine = create_engine(val)
        if self.is_sqlite:
            import sqlite3
            sqlite3.register_adapter(str, lambda s: s.decode('utf8'))
        session_factory = sessionmaker(self.engine)
        self.session = scoped_session(session_factory)

    @property
    def is_postgresql(self):
        return 'postgres' in self.dialect_name

    @property
    def is_sqlite(self):
        return 'sqlite' in self.dialect_name

    @classmethod
    def factory(cls, **kwargs):
        """ helper method to open a Database object (and optionally create the tables in that Database) """
        db = cls(**kwargs)
        if kwargs.get('create'):
            db.create()
        return db

    @classmethod
    def factory_from_cmdline(cls, args):
        """ helper method to open a Database via a set of cmdline args object """
        kwargs = vars(args)
        return cls.factory(**kwargs)

    def prep_an_orm_class(self, orm_cls):
        self.prep_orm_class(orm_cls, self.schema, self.is_geospatial)

    @classmethod
    def prep_orm_class(cls, orm_cls, schema=None, is_geospatial=False):
        """
        helper method to ready an ORM class (see Base and it's children) according to this Database's settings
        :why?: sometimes you might have classes you want as part of a query, but you don't want those classes
        available in the Database.classes() or Database.sorted_classes(), since these tables are not being loaded, etc..
        """
        if is_geospatial and hasattr(orm_cls, 'add_geometry_column'):
            orm_cls.add_geometry_column()

        if schema:
            orm_cls.set_schema(schema)

    @classmethod
    def prep_gtfsdb_model_classes(cls, schema=None, is_geo=False):
        for c in cls.get_base_subclasses():
            cls.prep_orm_class(c, schema, is_geo)

    @contextmanager
    def managed_session(self, *args, **kwds):
        """
        will return a session that you can use w/in a 'with' statement
        :see https://docs.python.org/3/library/contextlib.html#utilities :
        """
        log.debug("get managed session")
        session = self.session()
        try:
            yield session
        finally:
            log.debug("close managed session")
            session.close()



import logging
import time

from geoalchemy2 import Geometry
from sqlalchemy import Column, Integer, Numeric, String
from sqlalchemy.orm import deferred, relationship
from sqlalchemy.sql import func

from gtfsdb import config
from gtfsdb.model.base import Base


__all__ = ['Pattern', 'Shape']


log = logging.getLogger(__name__)


class Pattern(Base):
    datasource = config.DATASOURCE_DERIVED

    __tablename__ = 'patterns'

    shape_id = Column(String(255), primary_key=True, index=True)
    pattern_dist = Column(Numeric(20, 10))

    trips = relationship(
        'Trip',
        primaryjoin='Pattern.shape_id==Trip.shape_id',
        foreign_keys='(Pattern.shape_id)',
        uselist=True, viewonly=True)

    shapes = relationship(
        'Shape',
        primaryjoin='Pattern.shape_id==Shape.shape_id',
        foreign_keys='(Shape.shape_id)',
        uselist=True, viewonly=True)

    @classmethod
    def add_geometry_column(cls):
        if not hasattr(cls, 'geom'):
            cls.geom = deferred(Column(Geometry(geometry_type='LINESTRING', srid=config.SRID)))

    def geom_from_shape(self, points):
        coords = ['{0} {1}'.format(r.shape_pt_lon, r.shape_pt_lat) for r in points]
        self.geom = 'SRID={0};LINESTRING({1})'.format(config.SRID, ','.join(coords))

    @classmethod
    def load(cls, db, **kwargs):
        start_time = time.time()
        session = db.session
        q = session.query(
            Shape.shape_id,
            func.max(Shape.shape_dist_traveled).label('dist')
        )
        shapes = q.group_by(Shape.shape_id)
        for shape in shapes:
            pattern = cls()
            pattern.shape_id = shape.shape_id
            pattern.pattern_dist = shape.dist
            if hasattr(cls, 'geom'):
                q = session.query(Shape)
                q = q.filter(Shape.shape_id == shape.shape_id)
                q = q.order_by(Shape.shape_pt_sequence)
                pattern.geom_from_shape(q)
            session.add(pattern)
        session.commit()
        session.close()
        processing_time = time.time() - start_time
        log.debug('{0}.load ({1:.0f} seconds)'.format(
            cls.__name__, processing_time))


class Shape(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'shapes.txt'

    __tablename__ = 'shapes'

    shape_id = Column(String(255), primary_key=True, index=True)
    shape_pt_lat = Column(Numeric(12, 9))
    shape_pt_lon = Column(Numeric(12, 9))
    shape_pt_sequence = Column(Integer, primary_key=True, index=True)
    shape_dist_traveled = Column(Numeric(20, 10))

    @classmethod
    def add_geometry_column(cls):
        if not hasattr(cls, 'geom'):
            cls.geom = Column(Geometry(geometry_type='POINT', srid=config.SRID))

    @classmethod
    def add_geom_to_dict(cls, row):
        args = (config.SRID, row['shape_pt_lon'], row['shape_pt_lat'])
        row['geom'] = 'SRID={0};POINT({1} {2})'.format(*args)



from gtfsdb.model.db import Database
from gtfsdb.model.gtfs import GTFS

from gtfsdb.model.agency import Agency  # noqa
from gtfsdb.model.calendar import *  # noqa
from gtfsdb.model.fare import *  # noqa
from gtfsdb.model.feed_info import FeedInfo  # noqa
from gtfsdb.model.frequency import Frequency  # noqa
from gtfsdb.model.route import *  # noqa
from gtfsdb.model.route_stop import *  # noqa
from gtfsdb.model.shape import *  # noqa
from gtfsdb.model.stop import *  # noqa
from gtfsdb.model.stop_feature import *  # noqa
from gtfsdb.model.stop_time import StopTime  # noqa
from gtfsdb.model.transfer import Transfer  # noqa
from gtfsdb.model.trip import Trip  # noqa
from gtfsdb.model.block import Block  # noqa


SORTED_CLASS_NAMES = [
    RouteType.__name__,
    RouteFilter.__name__,
    FeedInfo.__name__,
    Agency.__name__,
    Block.__name__,
    Calendar.__name__,
    CalendarDate.__name__,
    Route.__name__,
    RouteDirection.__name__,
    Stop.__name__,
    StopFeature.__name__,
    Transfer.__name__,
    Shape.__name__,
    Pattern.__name__,
    Trip.__name__,
    StopTime.__name__,
    RouteStop.__name__,
    Frequency.__name__,
    FareAttribute.__name__,
    FareRule.__name__,
    UniversalCalendar.__name__,
]


CURRENT_CLASS_NAMES = [
    CurrentRoutes.__name__,
    CurrentRouteStops.__name__,
    CurrentStops.__name__,
]


from collections import defaultdict

from sqlalchemy import Column, Integer, Numeric, String
from sqlalchemy.orm import joinedload, joinedload_all, object_session, relationship

from gtfsdb import config, util
from gtfsdb.model.base import Base
from .stop_base import StopBase

import logging
log = logging.getLogger(__name__)


class Stop(Base, StopBase):
    datasource = config.DATASOURCE_GTFS
    filename = 'stops.txt'

    __tablename__ = 'stops'

    stop_id = Column(String(255), primary_key=True, index=True, nullable=False)
    stop_code = Column(String(50))
    stop_name = Column(String(255), nullable=False)
    stop_desc = Column(String(255))
    stop_lat = Column(Numeric(12, 9), nullable=False)
    stop_lon = Column(Numeric(12, 9), nullable=False)
    zone_id = Column(String(50))
    stop_url = Column(String(255))
    location_type = Column(Integer, index=True, default=0)
    parent_station = Column(String(255))
    stop_timezone = Column(String(50))
    wheelchair_boarding = Column(Integer, default=0)
    platform_code = Column(String(50))
    direction = Column(String(50))
    position = Column(String(50))

    stop_features = relationship(
        'StopFeature',
        primaryjoin='Stop.stop_id==StopFeature.stop_id',
        foreign_keys='(Stop.stop_id)',
        uselist=True, viewonly=True)

    stop_times = relationship(
        'StopTime',
        primaryjoin='Stop.stop_id==StopTime.stop_id',
        foreign_keys='(Stop.stop_id)',
        uselist=True, viewonly=True)

    @classmethod
    def add_geom_to_dict(cls, row):
        point = util.Point.make_geo(row['stop_lon'], row['stop_lat'], config.SRID)
        row['geom'] = point

    @property
    def routes(self):
        """
        return list of routes servicing this stop
        @todo: rewrite the cache to use timeout checking in Base.py
        """
        try:
            self._routes
        except AttributeError:
            from gtfsdb.model.route import Route
            from gtfsdb.model.trip import Trip
            from gtfsdb.model.stop_time import StopTime
            session = object_session(self)
            q = session.query(Route)
            f = ((StopTime.stop_id == self.stop_id) & (StopTime.departure_time != ''))
            q = q.filter(Route.trips.any(Trip.stop_times.any(f)))
            q = q.order_by(Route.route_sort_order)
            self._routes = q.all()
        return self._routes

    @property
    def headsigns(self):
        """
        Returns a dictionary of all unique (route_id, headsign) tuples used
        at the stop and the number of trips the head sign is used
        """
        if not hasattr(self, '_headsigns'):
            from gtfsdb.model.stop_time import StopTime
            self._headsigns = defaultdict(int)
            session = object_session(self)
            log.info("QUERY StopTime")
            q = session.query(StopTime)
            q = q.options(joinedload_all('trip.route'))
            q = q.filter_by(stop_id=self.stop_id)
            for r in q:
                headsign = r.stop_headsign or r.trip.trip_headsign
                self._headsigns[(r.trip.route, headsign)] += 1
        return self._headsigns

    @property
    def agencies(self):
        """
        return list of agency ids with routes hitting this stop
        @todo: rewrite the cache to use timeout checking in Base.py
        """
        try:
            self._agencies
        except AttributeError:
            self._agencies = []
            if self.routes:
                for r in self.routes:
                    if r.agency_id not in self._agencies:
                        self.agencies.append(r.agency_id)
        return self._agencies

    @property
    def amenities(self):
        """
        return list of strings for the stop amenity (feature) names
        """
        try:
            self._amenities
        except AttributeError:
            self._amenities = []
            if self.stop_features and len(self.stop_features) > 0:
                for f in self.stop_features:
                    n = f.feature_name
                    if n and len(n) > 0:
                        self._amenities.append(n)
                self._amenities = sorted(self._amenities)
        return self._amenities

    def is_active(self, date=None):
        """
        :return False whenever we see that the stop has zero stop_times on the given input date
                (which defaults to 'today')

        @NOTE: use caution with this routine.  calling this for multiple stops can really slow things down,
               since you're querying large trip and stop_time tables, and asking for a schedule of each stop
               I used to call this multiple times via route_stop to make sure each stop was active ... that
               was really bad performance wise.
        """
        from gtfsdb.model.stop_time import StopTime

        # import pdb; pdb.set_trace()
        _is_active = False
        date = util.check_date(date)
        st = StopTime.get_departure_schedule(self.session, self.stop_id, date, limit=1)
        if st and len(st) > 0:
            _is_active = True
        return _is_active

    @classmethod
    def query_active_stops(cls, session, limit=None, location_types=[0], active_filter=True, date=None):
        """
        check for active stops
        """
        # step 1: get stops
        q = session.query(Stop)
        if limit:
            q = q.limit(limit)
        if location_types and len(location_types) > 0:
            # note: default is to filter location_type=0, which is just stops (not stations)
            q.filter(Stop.location_type.in_(location_types))
        stops = q.all()

        # step 2: filter active stops only ???
        if active_filter:
            ret_val = []
            for s in stops:
                if s.is_active(date):
                    ret_val.append(s)
        else:
            ret_val = stops
        return ret_val

    @classmethod
    def query_active_stop_ids(cls, session, limit=None, active_filter=True):
        """
        return an array of stop_id / agencies pairs
        {stop_id:'2112', agencies:['C-TRAN', 'TRIMET']}
        """
        ret_val = []
        stops = cls.query_active_stops(session, limit, active_filter)
        for s in stops:
            ret_val.append({"stop_id": s.stop_id, "agencies": s.agencies})
        return ret_val


class CurrentStops(Base, StopBase):
    """
    this table is (optionally) used as a view into the currently active routes
    it is pre-calculated to list routes that are currently running service
    (GTFS can have multiple instances of the same route, with different aspects like name and direction)
    """
    datasource = config.DATASOURCE_DERIVED
    __tablename__ = 'current_stops'

    route_short_names = Column(String(1023))
    route_type = Column(Integer)
    route_type_other = Column(Integer)
    route_mode = Column(String(255))

    stop_id = Column(String(255), primary_key=True, index=True, nullable=False)
    location_type = Column(Integer)
    stop_lat = Column(Numeric(12, 9), nullable=False)
    stop_lon = Column(Numeric(12, 9), nullable=False)

    stop = relationship(
        Stop.__name__,
        primaryjoin='CurrentStops.stop_id==Stop.stop_id',
        foreign_keys='(CurrentStops.stop_id)',
        uselist=False, viewonly=True,
        lazy="joined", innerjoin=True,
    )

    def __init__(self, stop, session):
        """
        create a CurrentStop record from a stop record
        :param stop:
        :param session:
        """
        self.stop_id = stop.stop_id
        self.location_type = stop.location_type
        self.stop_lon = stop.stop_lon
        self.stop_lat = stop.stop_lat

        # copy the stop geom to CurrentStops (if we're in is_geospatial mode)
        if hasattr(stop, 'geom') and hasattr(self, 'geom'):
            self.geom = util.Point.make_geo(stop.stop_lon, stop.stop_lat, config.SRID)

        # convoluted route type assignment ... handle conditon where multiple modes (limited to 2) serve same stop
        # import pdb; pdb.set_trace()
        from .route_stop import CurrentRouteStops
        rs_list = CurrentRouteStops.query_route_short_names(session, stop, filter_active=True)
        for rs in rs_list:
            type = rs.get('type')
            if self.route_mode is None:
                self.route_type = type.route_type
                self.route_mode = type.otp_type
            elif type.is_different_mode(self.route_type):
                if type.is_lower_priority(self.route_type):
                    self.route_type_other = self.route_type
                    self.route_type = type.route_type
                    self.route_mode = type.otp_type
                else:
                    self.route_type_other = type.route_type

        # route short names
        self.route_short_names = CurrentRouteStops.to_route_short_names_as_string(rs_list)

    @classmethod
    def post_process(cls, db, **kwargs):
        """
        will update the current 'view' of this data
        """
        session = db.session()
        try:
            session.query(CurrentStops).delete()

            # import pdb; pdb.set_trace()
            for s in Stop.query_active_stops(session):
                c = CurrentStops(s, session)
                session.add(c)

            session.commit()
            session.flush()
        except Exception as e:
            log.warning(e)
            session.rollback()
        finally:
            session.flush()
            session.close()


__all__ = [Stop.__name__, CurrentStops.__name__]



from __future__ import print_function

import argparse

from gtfsdb import config
from gtfsdb.model.base import Base
from gtfsdb.api import database_load


def get_args(prog_name='gtfsdb-load',):
    """
    database load command-line arg parser and help util...
    """
    tables = sorted([t.name for t in Base.metadata.sorted_tables])
    parser = argparse.ArgumentParser(
        prog=prog_name,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('file', help='URL or local path to GTFS zip FILE')
    parser.add_argument('--batch_size', '-b', type=int, default=config.DEFAULT_BATCH_SIZE,
                        help='BATCH SIZE to use for memory management')
    parser.add_argument('--database_url', '-d', default=config.DEFAULT_DATABASE_URL,
                        help='DATABASE URL with appropriate privileges')
    parser.add_argument('--is_geospatial', '-g', action='store_true',
                        default=config.DEFAULT_IS_GEOSPATIAL,
                        help='Database supports GEOSPATIAL functions')
    parser.add_argument('--schema', '-s', default=config.DEFAULT_SCHEMA,
                        help='Database SCHEMA name')
    parser.add_argument('--tables', choices=tables, default=None, nargs='*',
                        help='Limited list of TABLES to load, if blank, load all tables')
    parser.add_argument('--current_tables', '-ct', default=False, action='store_true',
                        help="create tables that represent 'current' service (e.g., views)")
    parser.add_argument('--ignore_postprocess', '-np', default=False, action='store_true',
                        help="don't run any postprocess model routines (will leave some tables empty ... but will load raw gtfs data)")
    parser.add_argument('--ignore_blocks', '-nb', default=False, action='store_true',
                        help="don't bother populating the derrived block table")
    args = parser.parse_args()

    kwargs = dict(
        # common cmd line items
        url=args.database_url,
        schema=args.schema,
        is_geospatial=args.is_geospatial,
        current_tables=args.current_tables,

        # less used params
        do_postprocess=not args.ignore_postprocess,
        ignore_blocks=args.ignore_blocks,
        tables=args.tables,
        batch_size=args.batch_size
    )
    return args, kwargs


def gtfsdb_load():
    args, kwargs = get_args()
    database_load(args.file, **kwargs)


def route_stop_load():
    """
    written as a test / debug method for RS table loader
    """
    from gtfsdb import Database, RouteStop
    kwargs = get_args()[1]
    db = Database(**kwargs)
    RouteStop.load(db, **kwargs)


def current_tables_load(**kwargs):
    """
    current table loader
    """
    from gtfsdb import Database, CurrentRoutes, CurrentStops, CurrentRouteStops
    db = Database(**kwargs)
    for cls in [CurrentRoutes, CurrentRouteStops, CurrentStops]:
        db.create_table(cls)
        cls.post_process(db, **kwargs)


def current_tables_cmdline():
    kwargs = get_args('gtfsdb-current-load')[1]
    current_tables_load(**kwargs)


def db_connect_tester():
    """
    simple routine to connect to an existing database and list a few stops
    bin/connect-tester --database_url sqlite:///gtfs.db _no_gtfs_zip_needed_
    """
    from gtfsdb import Database, Stop, Route, StopTime
    args, kwargs = get_args('connect-tester')
    db = Database(**kwargs)
    for s in db.session.query(Stop).limit(2):
        print(s.stop_name)
    for r in db.session.query(Route).limit(2):
        print(r.route_name)
    stop_times = StopTime.get_departure_schedule(db.session, stop_id='11411')
    for st in stop_times:
        print(st.get_direction_name())
        break



from sqlalchemy import Column, Sequence
from sqlalchemy.types import Integer, Numeric, String

from gtfsdb import config
from gtfsdb.model.base import Base


__all__ = ['FareAttribute', 'FareRule']


class FareAttribute(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'fare_attributes.txt'

    __tablename__ = 'fare_attributes'

    fare_id = Column(String(255), primary_key=True)
    price = Column(Numeric(10, 2), nullable=False)
    currency_type = Column(String(255), nullable=False)
    payment_method = Column(Integer, nullable=False)
    transfers = Column(Integer)
    transfer_duration = Column(Integer)
    agency_id = Column(String(255))


class FareRule(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'fare_rules.txt'

    __tablename__ = 'fare_rules'

    id = Column(Integer, Sequence(None, optional=True), primary_key=True)
    fare_id = Column(String(255), index=True, nullable=False)
    route_id = Column(String(255))
    origin_id = Column(String(255))
    destination_id = Column(String(255))
    contains_id = Column(String(255))
    service_id = Column(String(255))



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from gtfsdb import *
from .base import load_sqlite


class TestLoad(unittest.TestCase):
    db = None

    def setUp(self):
        if TestLoad.db is None:
            TestLoad.db = load_sqlite(gtfs_name='sample-feed.zip')
        self.db = TestLoad.db

    def test_database_load(self):
        self.assertTrue(len(self.db.session.query(Stop).all()) > 0)
        self.assertTrue(len(self.db.session.query(Route).all()) > 0)



import logging

from gtfsdb import config
from gtfsdb.model.base import Base
from sqlalchemy import Column
from sqlalchemy.orm import relationship
from sqlalchemy.types import Integer, String

log = logging.getLogger(__name__)


class Trip(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'trips.txt'

    __tablename__ = 'trips'

    trip_id = Column(String(255), primary_key=True, index=True, nullable=False)
    route_id = Column(String(255), index=True, nullable=False)
    service_id = Column(String(255), index=True, nullable=False)
    direction_id = Column(Integer, index=True)
    block_id = Column(String(255), index=True)
    shape_id = Column(String(255), index=True, nullable=True)
    trip_type = Column(String(255))

    trip_headsign = Column(String(255))
    trip_short_name = Column(String(255))
    bikes_allowed = Column(Integer, default=0)
    wheelchair_accessible = Column(Integer, default=0)

    pattern = relationship(
        'Pattern',
        primaryjoin='Trip.shape_id==Pattern.shape_id',
        foreign_keys='(Trip.shape_id)',
        uselist=False, viewonly=True)

    route = relationship(
        'Route',
        primaryjoin='Trip.route_id==Route.route_id',
        foreign_keys='(Trip.route_id)',
        uselist=False, viewonly=True)

    stop_times = relationship(
        'StopTime',
        primaryjoin='Trip.trip_id==StopTime.trip_id',
        foreign_keys='(Trip.trip_id)',
        order_by='StopTime.stop_sequence',
        uselist=True, viewonly=True)

    universal_calendar = relationship(
        'UniversalCalendar',
        primaryjoin='Trip.service_id==UniversalCalendar.service_id',
        foreign_keys='(Trip.service_id)',
        uselist=True, viewonly=True)

    @classmethod
    def post_process(cls, db, **kwargs):
        trips = db.session.query(Trip).all()
        for t in trips:
            if not t.is_valid:
                log.warn("invalid trip: {0} only has {1} stop_time record (i.e., maybe the stops are coded as "
                         "non-public, and thus their stop time records didn't make it into the gtfs)"
                         .format(t.trip_id, t.trip_len))

    @classmethod
    def query_trip(cls, session, trip_id):
        """ return a trip via trip_id """
        ret_val = session.query(Trip).filter(Trip.trip_id==trip_id).one()
        return ret_val

    @property
    def start_stop(self):
        return self.stop_times[0].stop

    @property
    def end_stop(self):
        return self.stop_times[-1].stop

    @property
    def start_time(self):
        return self.stop_times[0].departure_time

    @property
    def end_time(self):
        return self.stop_times[-1].arrival_time

    @property
    def trip_len(self):
        ret_val = 0
        if self.stop_times:
            ret_val = len(self.stop_times)
        return ret_val

    @property
    def is_valid(self):
        # trip has to have multiple stop times to be valid, else it's not a trip...
        return self.trip_len >= 2



from sqlalchemy import Column
from sqlalchemy.types import Date, String

from gtfsdb import config
from gtfsdb.model.base import Base


class FeedInfo(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'feed_info.txt'

    __tablename__ = 'feed_info'

    feed_publisher_name = Column(String(255), primary_key=True)
    feed_publisher_url = Column(String(255), nullable=False)
    feed_lang = Column(String(255), nullable=False)
    feed_start_date = Column(Date)
    feed_end_date = Column(Date)
    feed_version = Column(String(255))
    feed_license = Column(String(255))



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from .base import check_counts
from gtfsdb import *

import logging
log = logging.getLogger(__name__)


class TestGeomQueries(unittest.TestCase):
    """
    load current tables:
    bin/gtfsdb-current-load -g -s trimet -d postgresql://ott@localhost:5432/ott x
    """
    db = None
    DO_PG = False

    def setUp(self):
        from .test_current import TestCurrent
        if TestCurrent.DO_PG and TestGeomQueries.db is None:
            self.DO_PG = True
            self.db = Database(url=TestCurrent.PG_URL, schema=TestCurrent.PG_SCHEMA, is_geospatial=True, current_tables=True)

    def test_nearest(self):
        if self.DO_PG:
            point = util.Point(lat=45.53, lon=-122.6664, srid="4326")
            curr_stops = CurrentStops.query_stops_via_point(self.db.session(), point)
            stops = Stop.query_stops_via_point(self.db.session(), point)
            self.assertTrue(check_counts(curr_stops, stops))

    def test_bbox(self):
        if self.DO_PG and False:
            #import pdb; pdb.set_trace()
            bbox = util.BBox(min_lat=45.530, max_lat=45.535, min_lon=-122.665, max_lon=-122.667, srid="4326")
            curr_stops = CurrentStops.query_stops_via_bbox(self.db.session, bbox)
            stops = Stop.query_stops_via_bbox(self.db.session, bbox)
            self.assertTrue(check_counts(curr_stops, stops))



from sqlalchemy import Column, Sequence
from sqlalchemy.types import Integer, String

from gtfsdb import config
from gtfsdb.model.base import Base


class Transfer(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'transfers.txt'

    __tablename__ = 'transfers'

    id = Column(Integer, Sequence(None, optional=True), primary_key=True)
    from_stop_id = Column(String(255))
    to_stop_id = Column(String(255))
    transfer_type = Column(Integer, index=True, default=0)
    min_transfer_time = Column(Integer)



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from .base import *
from gtfsdb import *
from gtfsdb import util


import logging
log = logging.getLogger(__name__)


class TestCurrent(unittest.TestCase):
    db = None
    DO_PG = False
    PG_URL = "postgresql://ott@localhost:5432/ott"
    PG_SCHEMA = "current_test"

    def setUp(self):
        if TestCurrent.db is None:
            TestCurrent.db = load_pgsql(self.PG_URL, self.PG_SCHEMA) if self.DO_PG else load_sqlite('curr')
        self.db = TestCurrent.db

    def check_query_counts(self, clz1, clz2):
        n1 = self.db.session.query(clz1).all()
        n2 = self.db.session.query(clz2).all()
        return check_counts(n1, n2)

    def test_load(self):
        self.assertTrue(self.check_query_counts(Stop,  CurrentStops))
        self.assertTrue(self.check_query_counts(Route, CurrentRoutes))
        self.assertTrue(self.check_query_counts(RouteStop, CurrentRouteStops))
        """""
        cr_list = self.db.session.query(CurrentRoutes).all()
        for cr in cr_list:
            self.assertTrue(cr.route is not None)
        """""

    def test_routes(self):
        routes = CurrentRoutes.query_active_routes(self.db.session())
        self.assertTrue(len(routes) > 0)

    def test_stops(self):
        stops = CurrentStops.query_stops(self.db.session(), limit=1)
        self.assertTrue(len(stops) == 1)

    def test_route_stops(self):
        #import pdb; pdb.set_trace()

        # DADAN has 2 routes active now
        routes = CurrentRouteStops.unique_routes_at_stop(self.db.session(), stop_id="DADAN")
        self.assertTrue(len(routes) == 2)
        self.assertTrue(routes[0].route_id in ('NEW', 'ALWAYS'))
        self.assertTrue(routes[1].route_id in ('NEW', 'ALWAYS'))

        # DADAN has 3 routes total ... RouteStop isn't filtering current stops, so will show older inactive route
        routes = RouteStop.unique_routes_at_stop(self.db.session(), stop_id="DADAN")
        self.assertTrue(len(routes) == 3)

        # OLD is not active, so CurrentStops should not have OLD as stop in the current route stop table
        routes = CurrentRouteStops.unique_routes_at_stop(self.db.session(), stop_id="OLD")
        self.assertTrue(len(routes) == 0)

        # although OLD is not active, RouteStop should show this route stop, since it's not filtering for is_active
        routes = RouteStop.unique_routes_at_stop(self.db.session(), stop_id="OLD")
        self.assertTrue(len(routes) == 1)
        self.assertTrue(routes[0].route_id == 'ALWAYS')

    def test_stops_point(self):
        if self.DO_PG:
            point = util.Point(lat=36.915, lon=-116.762, srid="4326")
            curr_stops = CurrentStops.query_stops_via_point(self.db.session(), point)
            stops = Stop.query_stops_via_point(self.db.session(), point)
            self.assertTrue(check_counts(curr_stops, stops))

    def test_stops_bbox(self):
        if self.DO_PG:
            bbox = util.BBox(min_lat=36.0, max_lat=37.0, min_lon=-117.5, max_lon=-116.0, srid="4326")
            curr_stops = CurrentStops.query_stops_via_bbox(self.db.session, bbox)
            stops = Stop.query_stops_via_bbox(self.db.session, bbox)
            self.assertTrue(check_counts(curr_stops, stops))



from sqlalchemy import Column
from sqlalchemy.orm import relationship
from sqlalchemy.types import Integer, String

from gtfsdb import config
from gtfsdb.model.base import Base


class Frequency(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'frequencies.txt'

    __tablename__ = 'frequencies'

    trip_id = Column(String(255), primary_key=True)
    start_time = Column(String(8), primary_key=True)
    end_time = Column(String(8))
    headway_secs = Column(Integer)
    exact_times = Column(Integer)

    trip = relationship(
        'Trip',
        primaryjoin='Frequency.trip_id==Trip.trip_id',
        foreign_keys='(Frequency.trip_id)',
        uselist=False, viewonly=True)



from gtfsdb import *
from .base import load_sqlite

try:
    import unittest2 as unittest
except ImportError:
    import unittest

import datetime
import logging
log = logging.getLogger(__name__)


class BasicModelTests(object):
    db = load_sqlite(gtfs_name='large-sample-feed.zip')

    def get_first(self):
        try:
            self._first
        except AttributeError:
            if hasattr(self, 'model'):
                self._first = self.db.session.query(self.model).first()
                return self._first

    def test_entity(self):
        if hasattr(self, 'model'):
            for r in self.db.session.query(self.model).limit(5):
                self.assertTrue(isinstance(r, self.model))


class TestRoute(unittest.TestCase, BasicModelTests):
    model = Route

    def test_dates(self):
        m = self.get_first()
        self.assertTrue(isinstance(m.start_date, datetime.date))
        self.assertTrue(isinstance(m.end_date, datetime.date))

    def test_active_date(self):
        # import pdb; pdb.set_trace()
        routes = Route.query_active_routes(self.db.session, datetime.date(2014, 6, 6))
        self.assertTrue(len(routes) > 1)
        for r in routes:
            self.assertTrue("good, I see active route id: {0}".format(r.route_id))

    def test_active_today(self):
        routes = Route.query_active_routes(self.db.session)
        for r in routes:
            self.assertFalse("we should not have any routes, but I see route id: {0}".format(r.route_id))


class TestRouteStop(unittest.TestCase, BasicModelTests):
    model = RouteStop

    def test_active_list(self):
        stops = RouteStop.query_active_stops(self.db.session, route_id="194", direction_id="1", date=datetime.date(2014, 6, 6))
        self.assertTrue(len(stops) > 1)
        for s in stops:
            self.assertTrue("good, I see active stop id: {0}".format(s.stop_id))

    def test_by_stop(self):
        stops = RouteStop.query_by_stop(self.db.session, stop_id="12883")
        self.assertTrue(len(stops) >= 1)

    def test_routes_serving_stop(self):
        routes = RouteStop.query_by_stop(self.db.session, stop_id="10767")
        self.assertTrue(len(routes) == 2)

    def test_route_stops(self):
        date = datetime.date(2014, 6, 6)
        stops = RouteStop.query_active_stops(self.db.session, route_id="193", direction_id="1", date=date)
        self.assertTrue(len(stops) > 5)

        stops = RouteStop.query_active_stops(self.db.session, route_id="193", direction_id="0", date=date)
        self.assertTrue(len(stops) > 5)

        stops = RouteStop.query_active_stops(self.db.session, route_id="194", direction_id="1", date=date)
        self.assertTrue(len(stops) > 5)

        stops = RouteStop.query_active_stops(self.db.session, route_id="194", direction_id="0", date=date)
        self.assertTrue(len(stops) > 5)

    def test_active_date(self):
        m = self.get_first()
        self.assertTrue(m.is_active(datetime.date(2014, 6, 6)))

    def test_active_today(self):
        m = self.get_first()
        self.assertFalse(m.is_active())


class TestRouteDirection(unittest.TestCase, BasicModelTests):
    model = RouteDirection


class TestTrip(unittest.TestCase, BasicModelTests):
    model = Trip

    def test_end_stop(self):
        m = self.get_first()
        self.assertTrue(isinstance(m.end_stop, Stop))

    def test_start_stop(self):
        m = self.get_first()
        self.assertTrue(isinstance(m.start_stop, Stop))

    def test_stop_times(self):
        m = self.get_first()
        for stop_time in m.stop_times:
            self.assertTrue(isinstance(stop_time, StopTime))

    def test_times(self):
        m = self.get_first()
        self.assertTrue(m.start_time)
        self.assertTrue(m.end_time)


class TestStop(unittest.TestCase, BasicModelTests):
    model = Stop

    def test_headsigns(self):
        m = self.get_first()
        self.assertTrue(isinstance(m.headsigns, dict))

    def test_routes(self):
        m = self.get_first()
        for r in m.routes:
            self.assertTrue(isinstance(r, Route))


class TestStopTimes(unittest.TestCase, BasicModelTests):
    model = StopTime


class TestAgency(unittest.TestCase, BasicModelTests):
    model = Agency


class TestCalendar(unittest.TestCase, BasicModelTests):
    model = Calendar


class TestCalendarDate(unittest.TestCase, BasicModelTests):
    model = CalendarDate


class TestFareAttribute(unittest.TestCase, BasicModelTests):
    model = FareAttribute


class TestFareRule(unittest.TestCase, BasicModelTests):
    model = FareRule


class TestShape(unittest.TestCase, BasicModelTests):
    model = Shape


class TestTransfer(unittest.TestCase, BasicModelTests):
    model = Transfer



import sys
import time

from sqlalchemy import Column, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.types import Integer, String

from gtfsdb import config
from gtfsdb.model.base import Base
from gtfsdb.model.trip import Trip

import logging
log = logging.getLogger(__name__)


class Block(Base):
    """
    This is really a BlockTripService table, in that we have entries for each Block / Trip pair, so that we can see
    the order of trips served by a given vehicle (block) for a particular service.

    One purpose is to know which trips might begin and end at a given stop .. we often don't want to show
    'arrival stops' in either our list of RouteStops or Stop Schedule listings...
    """
    datasource = config.DATASOURCE_DERIVED

    __tablename__ = 'blocks'

    id = Column(Integer, Sequence(None, optional=True), primary_key=True)
    sequence = Column(Integer)
    block_id = Column(String(255), index=True, nullable=False)
    service_id = Column(String(255), index=True, nullable=False)
    trip_id = Column(String(255), index=True, nullable=False)
    prev_trip_id = Column(String(255))
    next_trip_id = Column(String(255))
    start_stop_id = Column(String(255), index=True, nullable=False)
    end_stop_id = Column(String(255), index=True, nullable=False)

    universal_calendar = relationship(
        'UniversalCalendar',
        primaryjoin='Block.service_id==UniversalCalendar.service_id',
        foreign_keys='(Block.service_id)',
        uselist=True, viewonly=True)

    trip = relationship(
        'Trip',
        primaryjoin='Block.trip_id==Trip.trip_id',
        foreign_keys='(Block.trip_id)',
        uselist=False, viewonly=True)

    next_trip = relationship(
        'Trip',
        primaryjoin='Block.next_trip_id==Trip.trip_id',
        foreign_keys='(Block.next_trip_id)',
        uselist=False, viewonly=True)

    prev_trip = relationship(
        'Trip',
        primaryjoin='Block.prev_trip_id==Trip.trip_id',
        foreign_keys='(Block.prev_trip_id)',
        uselist=False, viewonly=True)

    start_stop = relationship(
        'Stop',
        primaryjoin='Stop.stop_id==Block.start_stop_id',
        foreign_keys='(Block.start_stop_id)',
        uselist=False, viewonly=True)

    end_stop = relationship(
        'Stop',
        primaryjoin='Stop.stop_id==Block.end_stop_id',
        foreign_keys='(Block.end_stop_id)',
        uselist=False, viewonly=True)

    def __init__(self, sequence, block_id, service_id, trip_id, prev_trip_id, next_trip_id, start_stop_id, end_stop_id):
        self.sequence = sequence
        self.block_id = block_id
        self.service_id = service_id
        self.trip_id = trip_id
        self.prev_trip_id = prev_trip_id
        self.next_trip_id = next_trip_id
        self.start_stop_id = start_stop_id
        self.end_stop_id = end_stop_id

    def is_arrival(self, stop_id=None):
        """
        check whether two sequential trips running on this block first arrive and then depart at this stop...
        if this is an 'arrival' stop, then we probably don't want to show it, etc...
        """
        ret_val = False

        # default is end_stop_id
        if stop_id is None:
            stop_id = self.end_stop_id

        if self.next_trip and self.next_trip.start_stop.stop_id == stop_id:
            # import pdb; pdb.set_trace()
            ret_val = True
        return ret_val

    @classmethod
    def load(cls, db, **kwargs):
        log.debug('{0}.load (loaded later in post_process)'.format(cls.__name__))
        pass

    @classmethod
    def post_process(cls, db, **kwargs):
        ignore_blocks = kwargs.get('ignore_blocks', None)
        log.debug('{0} {1}.post_process'.format("skip" if ignore_blocks else "run", cls.__name__))
        if not ignore_blocks:
            cls.populate(db)

    @classmethod
    def populate(cls, db):
        """
        loop thru a full trip table and break things into buckets based on service key and block id
        """
        start_time = time.time()
        batch_size = config.DEFAULT_BATCH_SIZE
        num_recs = 0

        # step 1: loop thru all trips, sorted by block and service key
        trips = db.session.query(Trip).order_by(Trip.block_id, Trip.service_id).all()
        i = 0
        while i < len(trips):
            # make sure the trip has a couple stops
            if not trips[i].is_valid:
                i = i + 1
                continue

            b = trips[i].block_id
            s = trips[i].service_id

            # need block (optional) and service id info ... if we don't have that, continue to next trip
            if b is None or s is None:
                i = i + 1
                continue

            # step 2: grab a batch of trips that have the same block and service id
            t = []
            while i < len(trips):
                if not trips[i].is_valid:
                    i = i + 1
                    continue

                if trips[i].block_id != b or trips[i].service_id != s:
                    break
                t.append(trips[i])
                i = i + 1

            # step 3: sort our bucket
            sorted_blocks = sorted(t, key=lambda t: t.start_time)
            sb_len = len(sorted_blocks) - 1

            # step 4: create block objects
            for j, k in enumerate(sorted_blocks):
                prev = None
                next = None
                if j > 0:
                    prev = sorted_blocks[j - 1].trip_id
                if j < sb_len:
                    next = sorted_blocks[j + 1].trip_id
                block = Block(
                    sequence=j + 1,
                    block_id=b,
                    service_id=s,
                    trip_id=k.trip_id,
                    prev_trip_id=prev,
                    next_trip_id=next,
                    start_stop_id=k.start_stop.stop_id,
                    end_stop_id=k.end_stop.stop_id
                )
                db.session.add(block)

            # step 5: insert in the db
            num_recs = num_recs + sb_len
            if num_recs >= batch_size:
                sys.stdout.write('*')
                db.session.flush()
                db.session.commit()
                num_recs = 0

        # step 5b: (final) insert into the db
        db.session.flush()
        db.session.commit()

        processing_time = time.time() - start_time
        log.debug('{0}.populate ({1:.0f} seconds)'.format(cls.__name__, processing_time))

    @classmethod
    def start_stop_ids(cls, session):
        """
        return an array of distinct starting stop_ids
        """
        ret_val = []
        blocks = session.query(Block).all()
        for b in blocks:
            if b.start_stop_id not in ret_val:
                ret_val.append(b.start_stop_id)
        return ret_val

    @classmethod
    def end_stop_ids(cls, session):
        """
        return an array of distinct ending stop_ids
        """
        ret_val = []
        blocks = session.query(Block).all()
        for b in blocks:
            if b.end_stop_id not in ret_val:
                ret_val.append(b.end_stop_id)
        return ret_val

    @classmethod
    def active_stop_ids(cls, session, limit=None):
        """
        return an array of unique starting and ending stop_ids
        use the dict {'stop_id':id} format for return (compatible with Stops.active_stop_ids())
        """
        stops = cls.start_stop_ids(session)
        stops.extend(cls.end_stop_ids(session))
        unique = set(stops)

        ret_val = []
        for i, s in enumerate(unique):
            if limit and i > int(limit):
                break
            ret_val.append({'stop_id': s})
        return ret_val

    @classmethod
    def blocks_by_stop_id(cls, session, stop_id, trip_id=None, service_keys=None, by_start_stop=False, by_end_stop=False):
        """
        query blocks by stop id and service keys ...
        """
        q = session.query(Block)
        if trip_id:
            q = q.filter(Block.trip_id == trip_id)
        if by_start_stop:
            q = q.filter(Block.start_stop_id == stop_id)
        if by_end_stop:
            q = q.filter(Block.end_stop_id == stop_id)
        if service_keys:
            q = q.filter(Block.service_id.in_(service_keys))
        blocks = q.all()
        return blocks

    @classmethod
    def blocks_by_start_stop_id(cls, session, stop_id, trip_id=None, service_keys=None):
        """
        query blocks by the start stop
        """
        return cls.blocks_by_stop_id(session, stop_id, trip_id=trip_id, service_keys=service_keys, by_start_stop=True)

    @classmethod
    def blocks_by_end_stop_id(cls, session, stop_id, trip_id=None, service_keys=None):
        """
        query blocks by the end stop
        """
        return cls.blocks_by_stop_id(session, stop_id, trip_id=trip_id, service_keys=service_keys, by_end_stop=True)

    @classmethod
    def blocks_by_trip_stop(cls, session, trip_id, stop_id, by_end_stop=True):
        """
        query blocks by the end stop
        """
        if by_end_stop:
            blocks = cls.blocks_by_end_stop_id(session, stop_id, trip_id=trip_id)
        else:
            blocks = cls.blocks_by_start_stop_id(session, stop_id, trip_id=trip_id)

        return blocks



import datetime
import logging

from gtfsdb import config
from gtfsdb.model.base import Base
from sqlalchemy import Column
from sqlalchemy.orm import joinedload_all, relationship
from sqlalchemy.sql.expression import func
from sqlalchemy.types import SmallInteger, Integer, Numeric, String

log = logging.getLogger(__name__)


class StopTime(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'stop_times.txt'

    __tablename__ = 'stop_times'

    trip_id = Column(String(255), primary_key=True, index=True, nullable=False)
    stop_id = Column(String(255), index=True, nullable=False)
    stop_sequence = Column(Integer, primary_key=True, nullable=False)
    arrival_time = Column(String(9))
    departure_time = Column(String(9), index=True)
    stop_headsign = Column(String(255))
    pickup_type = Column(Integer, default=0)
    drop_off_type = Column(Integer, default=0)
    shape_dist_traveled = Column(Numeric(20, 10))
    timepoint = Column(SmallInteger, index=True, default=0)

    stop = relationship(
        'Stop',
        primaryjoin='Stop.stop_id==StopTime.stop_id',
        foreign_keys='(StopTime.stop_id)',
        uselist=False, viewonly=True)

    trip = relationship(
        'Trip',
        primaryjoin='Trip.trip_id==StopTime.trip_id',
        foreign_keys='(StopTime.trip_id)',
        uselist=False, viewonly=True)

    def __init__(self, *args, **kwargs):
        super(StopTime, self).__init__(*args, **kwargs)
        if 'timepoint' not in kwargs:      # this logic is the int() equal of what was after changing this to SmallInt,
            if 'arrival_time' in kwargs:   # but I'm wondering if this is now going to set all stop times to timepoint=1=True ???
                self.timepoint = 1

    def get_headsign(self):
        """
        get the headsign at this stop ... rule is that if stop is empty, use trip headsign
        """
        ret_val = self.stop_headsign
        if not ret_val:
            ret_val = self.trip.trip_headsign
        return ret_val

    def get_direction_name(self, def_val="", banned=['Shuttle', 'MAX Shuttle', 'Garage', 'Center Garage', 'Merlo Garage', 'Powell Garage']):
        """
        returns either the headsign (priority) or the route direction name (when banned)
        (as long as one of these names are not banned and not the same name as the route name)
        """
        ret_val = def_val
        try:
            # step 0: create a banned list with the addition of our route_long_name
            banned = banned + [self.trip.route.route_long_name]

            headsign = self.get_headsign()
            if headsign and not any([headsign in s for s in banned]):
                # step 1: use the headsign as the direction name, just as long as the headsign is
                #         not null and not the same as the route name
                ret_val = headsign
            else:
                # step 2: lets use the direction name, if available
                d = self.trip.route.directions[self.trip.direction_id]
                if d.direction_name and not any([d.direction_name in s for s in banned]):
                    ret_val = d.direction_name.lstrip('to ').lstrip('To ')
        except Exception as e:
            log.debug(e)
            pass
        return ret_val

    def is_boarding_stop(self):
        """
        return whether the vehicle that is stopping at this stop, and at this time, is an
        in-revenue vehicle that a customer can actually board...

        pickup_type = 1 - No pickup available

        departure_time = None

        NOTE: in gtfsdb, we NULL out the departure times when the vehicle doesn't
              pick up anyone (e.g., at route end points, there are no departures...)

        @see: https://developers.google.com/transit/gtfs/reference#stop_times_fields
        """
        ret_val = True
        if self.pickup_type == 1 or self.departure_time is None:
            ret_val = False
        return ret_val

    @classmethod
    def post_process(cls, db, **kwargs):
        log.debug('{0}.post_process'.format(cls.__name__))
        # cls.null_out_last_stop_departures(db) ## commented out due to other processes
        pass

    @classmethod
    def null_out_last_stop_departures(cls, db):
        """
        delete all 'depature_time' values that appear for the last stop
        time of a given trip (e.g., the trip ends there, so there isn't a
        further vehicle departure / customer pickup for that stop time / trip pair)...

        -- query below shows null'd out stop times
        select * from ott.stop_times
        where COALESCE(arrival_time,'')='' or COALESCE(departure_time,'')=''

        NOTE: we know this breaks the current GTFS spec, which states that departure &
              arrival times must both exist for every stop time.  Sadly, GTFS is kinda wrong...
        """
        # step 1: remove the departure times at the end of a trip
        log.info("QUERY StopTime for all trip end times")
        sq = db.session.query(StopTime.trip_id, func.max(StopTime.stop_sequence).label('end_sequence'))
        sq = sq.group_by(StopTime.trip_id).subquery()
        q = db.session.query(StopTime)
        q = q.filter_by(trip_id=sq.c.trip_id, stop_sequence=sq.c.end_sequence)
        for st in q:
            if st.pickup_type == 1:
                st.departure_time = None

        # remove the arrival times at the start of a trip
        log.info("QUERY StopTime for all trip start times")
        sq = db.session.query(StopTime.trip_id, func.min(StopTime.stop_sequence).label('start_sequence'))
        sq = sq.group_by(StopTime.trip_id).subquery()
        q = db.session.query(StopTime)
        q = q.filter_by(trip_id=sq.c.trip_id, stop_sequence=sq.c.start_sequence)
        for st in q:
            if st.drop_off_type == 1:
                st.arrival_time = None

        db.session.flush()
        db.session.commit()
        db.session.close()

    @classmethod
    def get_service_keys_from_list(cls, stop_times):
        ret_val = []
        for s in stop_times:
            k = s.trip.service_id
            if k not in ret_val:
                ret_val.append(k)
        return ret_val

    @classmethod
    def get_departure_schedule(cls, session, stop_id, date=None, route_id=None, limit=None):
        """
        helper routine which returns the stop schedule for a give date
        """
        from gtfsdb.model.trip import Trip

        # step 0: make sure we have a valid date
        if date is None:
            date = datetime.date.today()

        # step 1: get stop times based on date
        log.debug("QUERY StopTime")
        q = session.query(StopTime)
        q = q.filter_by(stop_id=stop_id)
        q = q.filter(StopTime.departure_time is not None)
        q = q.filter(StopTime.trip.has(Trip.universal_calendar.any(date=date)))

        # step 2: apply an optional route filter
        if route_id:
            q = q.filter(StopTime.trip.has(Trip.route_id == route_id))

        # step 3: options to speed up /q
        q = q.options(joinedload_all('trip'))

        # step 4: order the stop times
        if limit is None or limit > 1:
            q = q.order_by(StopTime.departure_time)

        # step 5: limit results
        if limit:
            q = q.limit(limit)

        stop_times = q.all()
        ret_val = cls.block_filter(session, stop_id, stop_times)

        return ret_val

    @classmethod
    def block_filter(cls, session, stop_id, stop_times):
        """
        we don't want to show stop times that are arrivals, so we look at the blocks and figure out whether
        the input stop is the ending stop, and that there's a next trip starting at this same stop.
        """
        ret_val = stop_times
        if stop_times and len(stop_times) > 1:
            from gtfsdb.model.block import Block
            keys = cls.get_service_keys_from_list(stop_times)
            blocks = Block.blocks_by_end_stop_id(session, stop_id, service_keys=keys)
            if blocks:
                ret_val = []
                for s in stop_times:
                    block = None
                    for b in blocks:
                        if s.trip_id == b.trip_id and s.trip.block_id == b.block_id:
                            block = b
                            blocks.remove(b)
                            break

                    if block is None:
                        ret_val.append(s)
                    elif not block.is_arrival(stop_id):
                            ret_val.append(s)
                            # @todo maybe monkey patch stop_time with block, so we know about last trip

                            # this is an arrival trip, and the next trip
                            # (don't return the stop_time as a departure)
                            # this is the last trip of the day (so return it)
        return ret_val



from gtfsdb import Database, GTFS


def database_load(filename, **kwargs):
    """
    Basic API to load a GTFS zip file into a database

    arguments:
        filename: URL or local path to GTFS zip file

    keyword arguments:
        batch_size: record batch size for memory management
        is_geospatial: if database is support geo functions
        schema: database schema name
        tables: limited list of tables to load
        url: SQLAlchemy database url
    """
    db = Database(**kwargs)
    db.create()
    gtfs = GTFS(filename)
    gtfs.load(db, **kwargs)
    return db



from gtfsdb import util
import logging
log = logging.getLogger(__name__)


class RouteStopBase(object):

    @classmethod
    def is_arrival(cls, session, trip_id, stop_id):
        """
        :return True if it looks like this Trip / Stop pair is an arrival only
        NOTE: this routine might be EXPENSIVE since it is
        Further, this routine isn't well thought out...not sure block.is_arrival() works
        """
        _is_arrival = False

        from gtfsdb import Block
        blocks = Block.blocks_by_trip_stop(session, trip_id, stop_id)
        if blocks:
            for b in blocks:
                if b.is_arrival():
                    _is_arrival = True
                    break
        return _is_arrival


    @classmethod
    def query_route_short_names(cls, session, stop, filter_active=False):
        """
        :return an array of short names and types
        """
        from .route_stop import RouteStop

        # import pdb; pdb.set_trace()
        # step 1: create a short_names list
        short_names = []

        # step 2: use either route-dao list or find the active stops
        routes = stop.routes
        if routes is None or len(routes) == 0:
            routes = RouteStop.active_unique_routes_at_stop(session, stop_id=stop.stop_id)
            routes.sort(key=lambda x: x.route_sort_order, reverse=False)

        # step 3: build the short names list
        for r in routes:
            if filter_active and r.is_active() is False:
                continue
            sn = {'route_id': r.route_id, 'type': r.type, 'route_type': r.type.route_type, 'otp_type': r.type.otp_type, 'route_short_name': r.make_route_short_name(r)}
            short_names.append(sn)

        return short_names

    @classmethod
    def to_route_short_names_as_string(cls, short_names, sep=", "):
        """
        :return a string representing all short names (e.g., good for a tooltip on a stop popup)
        """
        ret_val = None
        for s in short_names:
            rsn = s.get('route_short_name')
            if rsn:
                if ret_val is None:
                    ret_val = rsn
                else:
                    ret_val = "{}{}{}".format(ret_val, sep, rsn)
        return ret_val

    @classmethod
    def query_by_stop(cls, session, stop_id, agency_id=None, date=None, count=None, sort=False):
        """
        get all route stop records by looking for a given stop_id.
        further filtering can be had by providing an active date and agency id
        """
        from .route_stop import RouteStop

        # step 1: query all route stops by stop id (and maybe agency)
        q = session.query(RouteStop).filter(RouteStop.stop_id == stop_id)
        if agency_id is not None:
            q = q.filter(RouteStop.agency_id == agency_id)

        # step 2: filter based on date
        if date:
            date = util.check_date(date)
            q = q.filter(RouteStop.start_date <= date).filter(date <= RouteStop.end_date)

        # step 3: sort the results based on order column
        if sort:
            q = q.order_by(RouteStop.order)

        # step 4: limit the number of objects returned by query
        if count:
            q = q.limit(count)

        ret_val = q.all()
        return ret_val

    @classmethod
    def unique_routes_at_stop(cls, session, stop_id, agency_id=None, date=None, route_name_filter=False):
        """
        get a unique set of route records by looking for a given stop_id.
        further filtering can be had by providing an active date and agency id, and route name
        """
        ret_val = []

        route_ids = []
        route_names = []

        route_stops = cls.query_by_stop(session, stop_id, agency_id, date, sort=True)
        for rs in route_stops:
            # step 1: filter(s) check against hashtable
            if rs.route_id in route_ids:
                continue
            if route_name_filter and rs.route.route_name in route_names:
                continue

            # step 2: add route attributes to cache hash-tables for later filtering (e.g. see filters above)
            route_ids.append(rs.route_id)
            route_names.append(rs.route.route_name)

            # step 3: this route is unique, so append route object to results
            ret_val.append(rs.route)
        return ret_val



import os
import sys
import datetime
import tempfile

import logging
log = logging.getLogger(__name__)


def get_all_subclasses(cls):
    """
    :see https://stackoverflow.com/questions/3862310/how-to-find-all-the-subclasses-of-a-class-given-its-name
    """
    ret_val = set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)]
    )
    return ret_val


def make_temp_sqlite_db_uri(name=None):
    """
    will return a FILE URI to a temp file, ala /tmp/bLaHh111 for the path of a new sqlite file db
    NOTE: name is optional ... if provided, the file will be named as such (good for testing and refreshing sqlite db)
    """
    if name:
        db_file = os.path.join(tempfile.gettempdir(), name)
    else:
        db_file = tempfile.mkstemp()[1]
    url = 'sqlite:///{0}'.format(db_file)
    log.debug("DATABASE TMP FILE: {0}".format(db_file))
    return url


def safe_get(obj, key, def_val=None):
    """
    try to return the key'd value from either a class or a dict
    (or return the raw value if we were handed a native type)
    """
    ret_val = def_val
    try:
        ret_val = getattr(obj, key)
    except:
        try:
            ret_val = obj[key]
        except:
            if isinstance(obj, (int, long, str)):
                ret_val = obj
    return ret_val


def safe_get_any(obj, keys, def_val=None):
    """
    :return object element value matching the first key to have an associated value
    """
    ret_val = def_val
    for k in keys:
        v = safe_get(obj, k)
        if v and len(v) > 0:
            ret_val = v
            break
    return ret_val


def check_date(in_date, fmt_list=['%Y-%m-%d', '%m/%d/%Y', '%m-%d-%Y'], def_val=None):
    """
    utility function to parse a request object for something that looks like a date object...
    """
    if def_val is None:
        def_val = datetime.date.today()

    if in_date is None:
        ret_val = def_val
    elif isinstance(in_date, datetime.date) or isinstance(in_date, datetime.datetime):
        ret_val = in_date
    else:
        ret_val = def_val
        for fmt in fmt_list:
            try:
                d = datetime.datetime.strptime(in_date, fmt).date()
                if d is not None:
                    ret_val = d
                    break
            except Exception as e:
                log.debug(e)
    return ret_val


class Point(object):
    is_valid = False

    def __init__(self, **kwargs):
        self.srid = kwargs.get('srid', None)
        try:
            self.lat = float(kwargs.get('lat'))
            self.lon = float(kwargs.get('lon'))
            self.is_valid = True
        except:
            self.lat = self.lon = None

    def get_point(self):
        return self.lon, self.lat

    def to_geojson(self):
        point = self.make_geo(self.lon, self.lat, self.srid)
        return point

    @classmethod
    def make_geo(cls, lon, lat, srid=None):
        geo = 'POINT({0} {1})'.format(lon, lat)
        if geo:
            geo = 'SRID={0};{1}'.format(srid, geo)
        return geo


class BBox(object):
    is_valid = False

    def __init__(self, **kwargs):
        self.srid = kwargs.get('srid', None)
        try:
            self.min_lat = float(kwargs.get('min_lat'))
            self.min_lon = float(kwargs.get('min_lon'))
            self.max_lat = float(kwargs.get('max_lat'))
            self.max_lon = float(kwargs.get('max_lon'))
            self.is_valid = True
        except:
            self.min_lat = self.min_lon = self.max_lat = self.max_lon = None

    def get_bbox(self):
        return self.min_lon, self.min_lat, self.max_lon, self.max_lat

    def to_geojson(self):
        poly = self.make_geo(self.min_lon, self.max_lon, self.min_lat, self.max_lat, self.srid)
        return poly

    @classmethod
    def make_geo(cls, left_lon, right_lon, bot_lat, top_lat, srid=None):
        """
        see: https://gis.stackexchange.com/questions/25797/select-bounding-box-using-postgis
        note: 5-pt POLY top-left, top-right, bot-right, bot-left,         ulx uly
                        llon/tlat, rlon/tlat, rlon/blat, min-lon/max-lat, min-lon/max-lat
        """
        geo = 'POLYGON(({0} {3}, {1} {3}, {1} {2}, {0} {2}, {0} {3}))'.format(left_lon, right_lon, bot_lat, top_lat)
        if geo:
            geo = 'SRID={0};{1}'.format(srid, geo)
        return geo


class UTF8Recoder(object):
    """Iterator that reads an encoded stream and encodes the input to UTF-8"""
    def __init__(self, f, encoding):
        import codecs
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        if sys.version_info >= (3, 0):
            return next(self.reader)
        else:
            return self.reader.next().encode('utf-8')

    def __next__(self):
        return self.next()



from sqlalchemy import Column, Sequence
from sqlalchemy.types import Integer, String

from gtfsdb import config
from gtfsdb.model.base import Base


__all__ = ['StopFeature']


class StopFeature(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'stop_features.txt'

    __tablename__ = 'stop_features'

    id = Column(Integer, Sequence(None, optional=True), primary_key=True)
    stop_id = Column(String(255), index=True, nullable=False)
    feature_type = Column(String(50), index=True, nullable=False)
    feature_name = Column(String(255))



import time

from gtfsdb import config, util
from gtfsdb.model.base import Base
from .route_base import RouteBase

from sqlalchemy import Column
from sqlalchemy.orm import deferred, relationship
from sqlalchemy.sql import func
from sqlalchemy.types import Integer, String

import logging
log = logging.getLogger(__name__)


class Route(Base, RouteBase):
    datasource = config.DATASOURCE_GTFS
    filename = 'routes.txt'

    __tablename__ = 'routes'

    route_id = Column(String(255), primary_key=True, index=True, nullable=False)
    agency_id = Column(String(255), index=True, nullable=True)
    route_short_name = Column(String(255))
    route_long_name = Column(String(255))
    route_desc = Column(String(1023))
    route_type = Column(Integer, index=True, nullable=False)
    route_url = Column(String(255))
    route_color = Column(String(6))
    route_text_color = Column(String(6))
    route_sort_order = Column(Integer, index=True)
    min_headway_minutes = Column(Integer)  # Trillium extension.

    agency = relationship(
        'Agency',
        primaryjoin='Route.agency_id==Agency.agency_id',
        foreign_keys='(Route.agency_id)',
        uselist=False, viewonly=True,
        lazy="joined", # don't innerjoin ... causes unit test errors
    )

    type = relationship(
        'RouteType',
        primaryjoin='Route.route_type==RouteType.route_type',
        foreign_keys='(Route.route_type)',
        uselist=False, viewonly=True,
        lazy="joined", innerjoin=True,
    )

    trips = relationship(
        'Trip',
        primaryjoin='Route.route_id==Trip.route_id',
        foreign_keys='(Route.route_id)',
        uselist=True, viewonly=True
    )

    directions = relationship(
        'RouteDirection',
        primaryjoin='Route.route_id==RouteDirection.route_id',
        foreign_keys='(Route.route_id)',
        uselist=True, viewonly=True
    )

    def is_active(self, date=None):
        """
        :return False whenever we see that the route start and end date are outside the
                input date (where the input date defaults to 'today')
        """
        _is_active = True
        if self.start_date or self.end_date:
            _is_active = False
            date = util.check_date(date)
            if self.start_date and self.end_date:  # keep this as nested if (don't combine due to below)
                if self.start_date <= date <= self.end_date:
                    _is_active = True
            elif self.start_date and self.start_date <= date:
                _is_active = True
            elif self.end_date and date <= self.end_date:
                _is_active = True
        return _is_active

    @property
    def route_name(self, fmt="{self.route_short_name}-{self.route_long_name}"):
        """
        build a route name out of long and short names...
        """
        if not self.is_cached_data_valid('_route_name'):
            log.warning("query route name")
            ret_val = self.route_long_name
            if self.route_long_name and self.route_short_name:
                ret_val = fmt.format(self=self)
            elif self.route_long_name is None:
                ret_val = self.route_short_name
            self._route_name = ret_val
            self.update_cached_data('_route_name')

        return self._route_name

    def direction_name(self, direction_id, def_val=''):
        ret_val = def_val
        try:
            dir = self.directions.filter(RouteDirection.direction_id == direction_id)
            if dir and dir.direction_name:
                ret_val = dir.direction_name
        except Exception as e:
            log.debug(e)
            pass
        return ret_val

    @property
    def start_date(self):
        return self._get_start_end_dates[0]

    @property
    def end_date(self):
        return self._get_start_end_dates[1]

    @property
    def _get_start_end_dates(self):
        """find the min & max date using Trip & UniversalCalendar"""
        if not self.is_cached_data_valid('_start_date'):
            from gtfsdb.model.calendar import UniversalCalendar
            q = self.session.query(func.min(UniversalCalendar.date), func.max(UniversalCalendar.date))
            q = q.filter(UniversalCalendar.trips.any(route_id=self.route_id))
            self._start_date, self._end_date = q.one()
            self.update_cached_data('_start_date')

        return self._start_date, self._end_date

    @classmethod
    def add_geometry_column(cls):
        if not hasattr(cls, 'geom'):
            from geoalchemy2 import Geometry
            cls.geom = deferred(Column(Geometry('MULTILINESTRING')))

    @classmethod
    def load_geoms(cls, db):
        """ load derived geometries, currently only written for PostgreSQL """
        from gtfsdb.model.shape import Pattern
        from gtfsdb.model.trip import Trip

        if db.is_geospatial and db.is_postgresql:
            start_time = time.time()
            session = db.session
            routes = session.query(Route).all()
            for route in routes:
                s = func.st_collect(Pattern.geom)
                s = func.st_multi(s)
                s = func.st_astext(s).label('geom')
                q = session.query(s)
                q = q.filter(Pattern.trips.any((Trip.route == route)))
                route.geom = q.first().geom
                session.merge(route)
            session.commit()
            processing_time = time.time() - start_time
            log.debug('{0}.load_geoms ({1:.0f} seconds)'.format(cls.__name__, processing_time))

    @classmethod
    def post_process(cls, db, **kwargs):
        log.debug('{0}.post_process'.format(cls.__name__))
        cls.load_geoms(db)


class CurrentRoutes(Base, RouteBase):
    """
    this table is (optionally) used as a view into the currently active routes
    it is pre-calculated to list routes that are currently running service
    (GTFS can have multiple instances of the same route, with different aspects like name and direction)
    """
    datasource = config.DATASOURCE_DERIVED
    __tablename__ = 'current_routes'

    route_id = Column(String(255), primary_key=True, index=True, nullable=False)
    route = relationship(
        Route.__name__,
        primaryjoin='CurrentRoutes.route_id==Route.route_id',
        foreign_keys='(CurrentRoutes.route_id)',
        uselist=False, viewonly=True,
        lazy="joined", innerjoin=True,
    )

    route_sort_order = Column(Integer)

    def __init__(self, route, def_order):
        self.route_id = route.route_id
        self.route_sort_order = route.route_sort_order if route.route_sort_order else def_order

    def is_active(self, date=None):
        ret_val = True
        if date:
            log.warning("you're calling CurrentRoutes.is_active with a date, which is both slow and redundant...")
            ret_val = self.route.is_active(date)
        return ret_val

    @classmethod
    def query_route(cls, session, route_id, detailed=False):
        """ get a gtfsdb Route object from the db """
        r = super(CurrentRoutes, cls).query_route(session, route_id, detailed)
        return r.route

    @classmethod
    def query_active_routes(cls, session, date=None):
        """
        wrap base active route query
        :return list of Route orm objects
        """
        ret_val = []
        if date:
            log.warning("you're calling CurrentRoutes.active_routes with a date, which is slow...")
            ret_val = Route.query_active_routes(session, date)
        else:
            try:
                clist = session.query(CurrentRoutes).order_by(CurrentRoutes.route_sort_order).all()
                for r in clist:
                    ret_val.append(r.route)
            except Exception as e:
                log.warning(e)
        return ret_val

    @classmethod
    def post_process(cls, db, **kwargs):
        """
        will update the current 'view' of this data

        steps:
          1. open transaction
          2. drop all data in this 'current' table
          3. select current routes as a list
          4. add those id's to this table
          5. other processing (cached results)
          6. commit
          7. close transaction
        """
        session = db.session()
        num_inserts = 0
        try:
            session.query(CurrentRoutes).delete()

            # import pdb; pdb.set_trace()
            rte_list = Route.query_active_routes(session)
            for i, r in enumerate(rte_list):
                c = CurrentRoutes(r, i+1)
                session.add(c)
                num_inserts += 1

            session.commit()
            session.flush()
        except Exception as e:
            log.warning(e)
            session.rollback()
        finally:
            session.flush()
            session.close()
        if num_inserts == 0:
            log.warning("CurrentRoutes did not insert any route records...hmmmm...")


class RouteDirection(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'route_directions.txt'

    __tablename__ = 'route_directions'

    route_id = Column(String(255), primary_key=True, index=True, nullable=False)
    direction_id = Column(Integer, primary_key=True, index=True, nullable=False)
    direction_name = Column(String(255))


class RouteType(Base):
    """
    OTP TYPES (come via service calls)
    0:TRAM, 1:SUBWAY, 2:RAIL, 3:BUS, 4:FERRY, 5:CABLE_CAR, 6:GONDOLA, 7:FUNICULAR
    :see https://github.com/opentripplanner/OpenTripPlanner/blob/master/src/main/java/org/opentripplanner/routing/core/TraverseMode.java :
    """
    datasource = config.DATASOURCE_LOOKUP
    filename = 'route_type.txt'
    __tablename__ = 'route_type'

    route_type = Column(Integer, primary_key=True, index=True, autoincrement=False)
    otp_type = Column(String(255))
    route_type_name = Column(String(255))
    route_type_desc = Column(String(1023))

    def is_bus(self):
        return self.route_type == 3

    def is_different_mode(self, cmp_route_type):
         return self.route_type != cmp_route_type

    def is_higher_priority(self, cmp_route_type):
        """ abitrary compare of route types, where lower numbrer means higher priority in terms mode ranking (sans bus) """
        ret_val = False
        if cmp_route_type != 3 and cmp_route_type < self.route_type:
            ret_val = True
        return ret_val

    def is_lower_priority(self, cmp_route_type):
        """ abitrary compare of route types, where lower numbrer means higher priority in terms mode ranking (sans bus) """
        ret_val = False
        if cmp_route_type != self.route_type:
            if cmp_route_type == 3 or cmp_route_type > self.route_type:
                ret_val = True
        return ret_val


class RouteFilter(Base):
    """
    list of filters to be used to cull routes from certain lists
    e.g., there might be Shuttles that you never want to be shown...you can load that data here, and
    use it in your queries
    """
    datasource = config.DATASOURCE_LOOKUP
    filename = 'route_filter.txt'
    __tablename__ = 'route_filters'

    route_id = Column(String(255), primary_key=True, index=True, nullable=False)
    agency_id = Column(String(255), index=True, nullable=True)
    description = Column(String(1023))


__all__ = [RouteType.__name__, Route.__name__, RouteDirection.__name__, RouteFilter.__name__, CurrentRoutes.__name__]


from sqlalchemy import Column, Integer, Numeric, String
from sqlalchemy.orm import joinedload, joinedload_all, object_session, relationship

from gtfsdb import config
from gtfsdb.util import BBox, Point

import logging
log = logging.getLogger(__name__)


class StopBase(object):
    """
    provides a generic set of stop query routines
    """

    def active_stops(self, date=None):
        """
        this common method will call route.is_active(), which means it will probably be slow
        note: this method, even when called from ActiveRoutes or ActiveStops, will probably be *slow*
        """
        ret_val = []
        try:
            for r in self.routes:
                if r.is_active(date):
                    ret_val.append(r)
        except Exception as e:
            log.warning(e)
        return ret_val

    @classmethod
    def add_geometry_column(cls):
        if not hasattr(cls, 'geom'):
            from geoalchemy2 import Geometry
            cls.geom = Column(Geometry(geometry_type='POINT', srid=config.SRID))

    @classmethod
    def query_orm_for_stop(cls, session, stop_id, detailed=False, agency=None):
        """
        simple utility for quering a stop from gtfsdb
        """
        ret_val = None
        try:
            log.info("query Stop for {}".format(stop_id))
            q = session.query(cls)
            q = q.filter(cls.stop_id == stop_id)
            # TODO q.filter(Stop.agency_id == agency_id)
            if detailed:
                try:
                    q = q.options(joinedload("stop_features"))
                except:
                    pass
            ret_val = q.one()
        except Exception as e:
            log.info(e)
        return ret_val

    @classmethod
    def generic_query_stops(cls, session, **kwargs):
        """
        query for list of this data
        """
        ret_val = []
        try:
            # import pdb; pdb.set_trace()
            clist = session.query(cls)
            limit = kwargs.get('limit')
            if limit:
                clist = clist.limit(limit)
            ret_val = clist.all()
        except Exception as e:
            log.warning(e)
        return ret_val

    @classmethod
    def query_stops_via_bbox(cls, session, bbox, limit=2000):
        ret_val = []
        try:
            log.info("query gtfsdb Stop table")
            q = session.query(cls)
            q = q.filter(cls.location_type == 0)  # just stops (not stations or entrances to stations)
            q = q.filter(cls.geom.ST_Within(bbox.to_geojson()))
            q = q.limit(limit + 10)
            ret_val = q.all()
        except Exception as e:
            log.warning(e)
        return ret_val

    @classmethod
    def query_stops_via_point(cls, session, point, limit=10):
        ret_val = []
        try:
            # import pdb; pdb.set_trace()
            log.info("query Stop table")
            q = session.query(cls)
            q = q.filter(cls.location_type == 0)
            q = q.order_by(cls.geom.distance_centroid(point.to_geojson()))
            q = q.limit(limit)
            ret_val = q.all()
        except Exception as e:
            log.warning(e)
        return ret_val

    @classmethod
    def query_stops(cls, session, **kwargs):
        """
        will query the db for a stop, either via bbox, point & distance or just an id
        :return list of stops
        """
        ret_val = []
        # import pdb; pdb.set_trace()
        if kwargs.get('lat') or kwargs.get('min_lat'):
            bbox = BBox(**kwargs)
            if bbox.is_valid:
                ret_val = cls.query_stops_via_bbox(session, bbox)
            else:
                point = Point(**kwargs)
                if point.is_valid:
                    ret_val = cls.query_stops_via_point_radius(session, point)
        else:
            ret_val = cls.generic_query_stops(session, **kwargs)
        return ret_val



import datetime
import logging
import time

from sqlalchemy import Column, Index
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.types import Date, SmallInteger, Integer, String

from gtfsdb import config
from gtfsdb.model.base import Base


__all__ = ['Calendar', 'CalendarDate', 'UniversalCalendar']


log = logging.getLogger(__name__)


class Calendar(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'calendar.txt'

    __tablename__ = 'calendar'
    __table_args__ = (Index('calendar_ix1', 'start_date', 'end_date'),)

    service_id = Column(String(255), primary_key=True, index=True, nullable=False)
    monday = Column(SmallInteger, nullable=False)
    tuesday = Column(SmallInteger, nullable=False)
    wednesday = Column(SmallInteger, nullable=False)
    thursday = Column(SmallInteger, nullable=False)
    friday = Column(SmallInteger, nullable=False)
    saturday = Column(SmallInteger, nullable=False)
    sunday = Column(SmallInteger, nullable=False)
    start_date = Column(Date, nullable=False)
    end_date = Column(Date, nullable=False)
    service_name = Column(String(255))  # Trillium extension, a human-readable name for the calendar.

    def weekday_list(self):
        weekday_dict = dict(monday=0, tuesday=1, wednesday=2, thursday=3, friday=4, saturday=5, sunday=6)
        item_func = weekday_dict.iteritems if hasattr(weekday_dict, 'iteritems') else weekday_dict.items
        return [v for k, v in item_func() if getattr(self, k)]

    def to_date_list(self):
        """
        TODO: we need better date limiting management here ... this routine could spin a long time w/forever dates
        TODO: for example, if the begin date is 1900 or end date is 9999, then that'll cause a major slowdown
        """
        date_list = []
        weekdays = self.weekday_list()
        diff = self.end_date - self.start_date
        for i in range(diff.days + 1):
            d = self.start_date + datetime.timedelta(days=i)
            if d.weekday() in weekdays:
                date_list.append(dict(service_id=self.service_id, date=d))
        return date_list


class CalendarDate(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'calendar_dates.txt'

    __tablename__ = 'calendar_dates'

    service_id = Column(String(255), primary_key=True, index=True, nullable=False)
    date = Column(Date, primary_key=True, index=True, nullable=False)
    exception_type = Column(Integer, nullable=False)

    @hybrid_property
    def is_addition(self):
        return self.exception_type == 1

    @hybrid_property
    def is_removal(self):
        return self.exception_type == 2


class UniversalCalendar(Base):
    datasource = config.DATASOURCE_DERIVED
    __tablename__ = 'universal_calendar'

    service_id = Column(String(255), primary_key=True, index=True, nullable=False)
    date = Column(Date, primary_key=True, index=True, nullable=False)

    trips = relationship(
        'Trip',
        primaryjoin='UniversalCalendar.service_id==Trip.service_id',
        foreign_keys='(UniversalCalendar.service_id)',
        uselist=True, viewonly=True)

    @classmethod
    def load(cls, db, **kwargs):
        start_time = time.time()
        session = db.session
        for c in session.query(Calendar):
            session.add_all([cls(**r) for r in c.to_date_list()])
        session.commit()
        q = session.query(CalendarDate)
        for calendar_date in q:
            cd_kwargs = dict(date=calendar_date.date,
                             service_id=calendar_date.service_id)
            if calendar_date.is_addition:
                session.merge(cls(**cd_kwargs))
            if calendar_date.is_removal:
                session.query(cls).filter_by(**cd_kwargs).delete()
        session.commit()
        process_time = time.time() - start_time
        log.debug('{0}.load ({1:.0f} seconds)'.format(cls.__name__, process_time))



try:
    import unittest2 as unittest
except ImportError:
    import unittest

import sys
import shutil
import datetime

from gtfsdb import *
from .base import load_sqlite

import logging
log = logging.getLogger(__name__)


class TestRouteStop(unittest.TestCase):
    model = RouteStop
    db = None

    def setUp(self):
        if TestRouteStop.db is None:
            TestRouteStop.db = load_sqlite()
        self.db = TestRouteStop.db

    def test_old_routes(self):
        date = datetime.date(2018, 12, 25)
        rs = RouteStop.query_active_stops(self.db.session, route_id="OLD", direction_id="1", date=date)
        self.assertTrue(len(rs) > 2)

        rs = RouteStop.query_active_stops(self.db.session, route_id="OLD", direction_id="0", date=date)
        self.assertTrue(len(rs) > 2)

        rs = RouteStop.query_active_stops(self.db.session, route_id="NEW", direction_id="1", date=date)
        self.assertTrue(len(rs) == 0)

        rs = RouteStop.query_active_stops(self.db.session, route_id="NEW", direction_id="0", date=date)
        self.assertTrue(len(rs) == 0)

    def test_via_stops(self):
        # date = datetime.date(2015, 6, 6)
        rs = RouteStop.query_by_stop(self.db.session, stop_id="OLD")
        self.assertTrue(len(rs) >= 2)

        routes = RouteStop.unique_routes_at_stop(self.db.session, stop_id="OLD")
        self.assertTrue(len(routes) == 1)

    def test_active_stop_list(self):
        rs = RouteStop.query_active_stops(self.db.session, route_id="ALWAYS", date=datetime.date(2015, 12, 25))
        self.assertTrue(len(rs) == 0)

        rs = RouteStop.query_active_stops(self.db.session, route_id="ALWAYS", date=datetime.date(2018, 12, 25))
        see_old_stop = False
        for r in rs:
            self.assertTrue(r.stop_id != "NEW")
            if r.stop_id == "OLD":
                see_old_stop = True
        self.assertTrue(see_old_stop)

        rs = RouteStop.query_active_stops(self.db.session, route_id="ALWAYS", date=datetime.date(2019, 1, 5))
        see_new_stop = False
        for r in rs:
            self.assertTrue(r.stop_id != "OLD")
            if r.stop_id == "NEW":
                see_new_stop = True
        self.assertTrue(see_new_stop)

    def test_stop_dates(self):
        active = RouteStop.is_stop_active(self.db.session, stop_id="OLD", date=datetime.date(2018, 12, 25))
        self.assertTrue(active)

        active = RouteStop.is_stop_active(self.db.session, stop_id="OLD", date=datetime.date(2019, 1, 3))
        self.assertFalse(active)

        active = RouteStop.is_stop_active(self.db.session, stop_id="OLD", date=datetime.date(2019, 6, 6))
        self.assertFalse(active)

        active = RouteStop.is_stop_active(self.db.session, stop_id="NEW", date=datetime.date(2018, 12, 25))
        self.assertFalse(active)

        active = RouteStop.is_stop_active(self.db.session, stop_id="NEW", date=datetime.date(2019, 6, 6))
        self.assertTrue(active)

    def test_new_routes(self):
        date = datetime.date(2019, 1, 1)
        rs = RouteStop.query_active_stops(self.db.session, route_id="NEW", direction_id="1", date=date)
        self.assertTrue(len(rs) > 2)

        rs = RouteStop.query_active_stops(self.db.session, route_id="NEW", direction_id="0", date=date)
        self.assertTrue(len(rs) > 2)

        rs = RouteStop.query_active_stops(self.db.session, route_id="OLD", direction_id="1", date=date)
        self.assertTrue(len(rs) == 0)

        rs = RouteStop.query_active_stops(self.db.session, route_id="OLD", direction_id="0", date=date)
        self.assertTrue(len(rs) == 0)

    def test_effective_dates(self):
        date = datetime.date(2019, 1, 1)
        rs = RouteStop.query_active_stops(self.db.session, route_id="NEW", direction_id="1", date=date)
        self.assertTrue(len(rs) > 2)

    def test_active_list(self):
        rs = RouteStop.query_active_stops(self.db.session, route_id="OLD", direction_id="1", date=datetime.date(2018, 12, 25))
        self.assertTrue(len(rs) > 1)
        for s in rs:
            self.assertTrue("good, I see active stop id: {0}".format(s.stop_id))


def main(argv):
    shutil.copyfile(TestRouteStop.db_file, "gtfs.db")
    t = TestRouteStop()
    t.test_active_list()

if __name__ == "__main__":
    main(sys.argv)



try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser
import logging.config
import os
from pkg_resources import resource_filename  # @UnresolvedImport


'''Parse configuration file and setup logging'''
config = ConfigParser()
ini_file = os.path.join(resource_filename('gtfsdb', 'configs'), 'app.ini')
config.read(ini_file)
if config.has_section('loggers'):
    logging.config.fileConfig(ini_file, disable_existing_loggers=False)


'''Application defaults'''
DEFAULT_BATCH_SIZE = 10000
DEFAULT_DATABASE_URL = 'sqlite://'
DEFAULT_IS_GEOSPATIAL = False
DEFAULT_SCHEMA = None

'''Data source constants'''
DATASOURCE_GTFS = 1
DATASOURCE_LOOKUP = 2
DATASOURCE_DERIVED = 3

'''Geometry constants'''
SRID = 4326



from contextlib import closing
import logging
import shutil
import tempfile
import time
try:
    from urllib import urlretrieve  # Python 2
except ImportError:
    from urllib.request import urlretrieve  # Python 3
import zipfile

from gtfsdb import config
from .route import Route

log = logging.getLogger(__name__)


class GTFS(object):

    def __init__(self, filename):
        self.file = filename
        try:
            self.local_file = urlretrieve(filename)[0]
        except Exception:
            try:
                self.local_file = urlretrieve("file:///" + filename)[0]
            except Exception:
                self.local_file = filename

    def load(self, db, **kwargs):
        """
        Load GTFS into database
        """
        # import pdb; pdb.set_trace()
        start_time = time.time()
        log.debug('GTFS.load: {0}'.format(self.file))

        # step 1: load .txt files from GTFS.zip, as well as derived tables & lookup tables from gtfsdb/data
        gtfs_directory = self.unzip()
        kwargs['gtfs_directory'] = gtfs_directory
        db.load_tables(**kwargs)
        shutil.rmtree(gtfs_directory)

        # step 2: call post process routines...
        db.postprocess_tables(**kwargs)

        # step 3: finish
        process_time = time.time() - start_time
        log.debug('GTFS.load ({0:.0f} seconds)'.format(process_time))

    def unzip(self, path=None):
        """
        Unzip GTFS files from URL/directory to path.
        """
        path = path if path else tempfile.mkdtemp()
        try:
            with closing(zipfile.ZipFile(self.local_file)) as z:
                z.extractall(path)
        except Exception as e:
            log.warning(e)
        return path



from gtfsdb import util

import logging
log = logging.getLogger(__name__)


class RouteBase(object):
    """
    provides a generic set of route query routines
    """

    def is_active(self, date=None):
        log.warning("calling abstract base class")
        return True

    @classmethod
    def query_route(cls, session, route_id, detailed=False):
        """
        simple utility for quering a route from gtfsdb
        """
        # import pdb; pdb.set_trace()
        ret_val = None
        try:
            log.info("query Route for {}".format(route_id))
            q = session.query(cls)
            q = q.filter(cls.route_id == route_id)
            if detailed:
                # todo: some joined junk, ala what we see in stops? -- q = q.options(joinedload("stop_features"))
                pass
            ret_val = q.one()
        except Exception as e:
            log.info(e)
        return ret_val

    @classmethod
    def query_route_list(cls, session):
        """
        :return list of *all* Route orm objects queried from the db
        """
        # import pdb; pdb.set_trace()
        from .route import RouteFilter
        routes = session.query(cls)\
            .filter(~cls.route_id.in_(session.query(RouteFilter.route_id)))\
            .order_by(cls.route_sort_order)\
            .all()
        return routes

    @classmethod
    def query_active_routes(cls, session, date=None):
        """
        :return list of *active* Route orm objects queried from the db
        :note 'active' is based on date ... this routine won't deal with holes in the
              schedule (e.g., when a route is not active for a period of time, due to construction)
        """
        # step 1: grab all routes
        routes = cls.query_route_list(session)

        # step 2: filter routes by active date
        ret_val = cls.filter_active_routes(routes, date)
        return ret_val

    @classmethod
    def filter_active_routes(cls, route_list, date=None):
        """
        filter an input list of route (orm) objects via is_active
        :return new list of routes filtered by date
        """
        # import pdb; pdb.set_trace()
        ret_val = []
        for r in route_list:
            if r and r.is_active(date):
                ret_val.append(r)
        return ret_val

    @classmethod
    def query_nearest_routes(cls, session, geom):
        """
        simple utility for quering a route from gtfsdb
        """
        ret_val = None

    @classmethod
    def query_active_route_ids(cls, session):
        """
        return an array of route_id / agency_id pairs
        {route_id:'2112', agency_id:'C-TRAN'}
        """
        ret_val = []
        routes = cls.query_active_routes(session)
        for r in routes:
            ret_val.append({"route_id": r.route_id, "agency_id": r.agency_id})
        return ret_val

    @classmethod
    def make_route_short_name(cls, route, def_name=None):
        """
        fix up the short name...
        """
        ret_val = def_name
        try:
            ret_val = util.safe_get_any(route, ['route_short_name', 'short_name', 'route_long_name', 'name'])

            # strip off 'Line' from last word, ala MAX Blue Line == MAX Blue
            if ret_val and ret_val.startswith('MAX') and ret_val.endswith('Line'):
                ret_val = " ".join(ret_val.split()[:-1])
            # special fix for Portland Streetcar
            if 'Portland Streetcar' in ret_val:
                ret_val = ret_val.replace('Portland Streetcar', 'PSC').strip()
            # fix WES
            if ret_val and ret_val.startswith('WES '):
                ret_val = "WES"
            # fix Portland Aerial Tram
            if ret_val and ret_val == 'Portland Aerial Tram':
                ret_val = "Aerial Tram"
        except Exception as e:
            log.warning(e)

        return ret_val




from setuptools import setup, find_packages
import sys


oracle_extras = ['cx_oracle>=5.1']
postgresql_extras = ['psycopg2>=2.4.2']
# dev_extras = oracle_extras + postgresql_extras
dev_extras = []

extras_require = dict(
    dev=dev_extras,
    oracle=oracle_extras,
    postgresql=postgresql_extras,
)

install_requires = [
    'geoalchemy2',
    'sqlalchemy',
    'psycopg2',
]

setup(
    name='gtfsdb',
    version='0.5.0',
    description='GTFS Database',
    long_description=open('README.rst').read(),
    keywords='GTFS',
    author="Open Transit Tools",
    author_email="info@opentransittools.org",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        'console_scripts': [
            'gtfsdb-load = gtfsdb.scripts:gtfsdb_load',
            'gtfsdb-current-load = gtfsdb.scripts:current_tables_cmdline',
            'rs-test = gtfsdb.scripts:route_stop_load',
            'connect-tester = gtfsdb.scripts:db_connect_tester',
        ]
    },
    classifiers=(
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ),
)



from sqlalchemy import Column, Sequence
from sqlalchemy.types import Integer, String

from gtfsdb import config
from gtfsdb.model.base import Base


class Agency(Base):
    datasource = config.DATASOURCE_GTFS
    filename = 'agency.txt'

    __tablename__ = 'agency'

    id = Column(Integer, Sequence(None, optional=True), primary_key=True, nullable=True)
    agency_id = Column(String(255), index=True, unique=True)
    agency_name = Column(String(255), nullable=False)
    agency_url = Column(String(255), nullable=False)
    agency_timezone = Column(String(50), nullable=False)
    agency_lang = Column(String(10))
    agency_phone = Column(String(50))
    agency_fare_url = Column(String(255))
    agency_email = Column(String(255))



from pkg_resources import resource_filename  # @UnresolvedImport

from gtfsdb import config, util
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import object_session

import csv
import datetime
import os
import sys
import time

import logging
log = logging.getLogger(__name__)


try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    str = str
    unicode = str
    bytes = bytes
    basestring = (str, bytes)
else:
    # 'unicode' exists, must be Python 2
    str = str
    unicode = unicode
    bytes = str
    basestring = basestring


class _Base(object):

    filename = None

    @property
    def session(self):
        #import pdb; pdb.set_trace()
        ret_val = None
        try:
            ret_val = object_session(self)
        except:
            log.warning("can't get a session from object")
        return ret_val

    @classmethod
    def set_schema(cls, schema):
        cls.__table__.schema = schema

    @classmethod
    def make_geom_lazy(cls):
        from sqlalchemy.orm import deferred
        try:
            cls.__mapper__.add_property('geom', deferred(cls.__table__.c.geom))
        except Exception as e:
            log.warning(e)

    @classmethod
    def from_dict(cls, attrs):
        clean_dict = cls.make_record(attrs)
        return cls(**clean_dict)

    @property
    def to_dict(self):
        """
        convert a SQLAlchemy object into a dict that is serializable to JSON
        """
        ret_val = self.__dict__.copy()

        """ not crazy about this hack, but ... the __dict__ on a SQLAlchemy
        object contains hidden crap that we delete from the class dict
        """
        if set(['_sa_instance_state']).issubset(ret_val):
            del ret_val['_sa_instance_state']

        """ we're using 'created' as the date parameter, so convert values
        to strings <TODO>: better would be to detect date & datetime objects,
        and convert those...
        """
        if set(['created']).issubset(ret_val):
            ret_val['created'] = ret_val['created'].__str__()

        return ret_val

    def get_up_date_name(self, attribute_name):
        """
        return attribute name of where we'll store an update variable
        """
        return "{0}_update_utc".format(attribute_name)

    def is_cached_data_valid(self, attribute_name, max_age=2):
        """
        we have to see both the attribute name exist in our object, as well as
        that object having a last update date (@see update_cached_data below)
        and that update date being less than 2 days ago...
        """
        ret_val = False
        try:
            # import pdb; pdb.set_trace()
            if hasattr(self, attribute_name):
                attribute_update = self.get_up_date_name(attribute_name)
                if hasattr(self, attribute_update):
                    epoch = datetime.datetime.utcfromtimestamp(0)
                    delta = getattr(self, attribute_update) - epoch
                    if delta.days <= max_age:
                        ret_val = True
        except:
            log.warning("is_cached_data_valid(): saw a cache exception with attribute {0}".format(attribute_name))
            ret_val = False

        return ret_val

    def update_cached_data(self, attribute_name):
        """
        """
        try:
            # import pdb; pdb.set_trace()
            attribute_update = self.get_up_date_name(attribute_name)
            setattr(self, attribute_update, datetime.datetime.now())
        except:
            log.warning("update_cached_data(): threw an exception with attribute {0}".format(attribute_name))

    @classmethod
    def load(cls, db, **kwargs):
        """
        Load method for ORM

        arguments:
            db: instance of gtfsdb.Database

        keyword arguments:
            gtfs_directory: path to unzipped GTFS files
            batch_size: batch size for memory management
        """

        # step 0: set up some vars, including setting the log output to show the child of base that we're processing
        start_time = time.time()
        batch_size = kwargs.get('batch_size', config.DEFAULT_BATCH_SIZE)
        log = logging.getLogger(cls.__module__)

        # step 1: check that we have elements of a file path (a file name and a directory) for the data we'll load
        if cls.filename is None:
            if cls.datasource is not config.DATASOURCE_DERIVED:
                log.info("{0} didn't specify a 'filename', so won't bohter trying to load() a null file (early exit from load()).".format(cls.__name__))
            return  # note early exit
        if cls.datasource is not config.DATASOURCE_GTFS and cls.datasource is not config.DATASOURCE_LOOKUP:
            log.info("{0}.datasource != DATASOURCE_GTFS or DATASOURCE_LOOKUP (exit load).".format(cls.__name__))
            return  # note early exit

        # step 2: load either a GTFS file from the unzipped file or a resource file (from a dir specified in config)
        directory = None
        if cls.datasource == config.DATASOURCE_GTFS:
            directory = kwargs.get('gtfs_directory')
        elif cls.datasource == config.DATASOURCE_LOOKUP:
            directory = resource_filename('gtfsdb', 'data')

        # step 3: load the file
        log.info("load {0}".format(cls.__name__))
        records = []
        file_path = os.path.join(directory, cls.filename)
        if os.path.exists(file_path):
            if sys.version_info >= (3, 0):
                f = open(file_path, 'rb')
            else:
                f = open(file_path, 'r')
            utf8_file = util.UTF8Recoder(f, 'utf-8-sig')
            reader = csv.DictReader(utf8_file)
            reader.fieldnames = [field.strip().lower() for field in reader.fieldnames]
            table = cls.__table__
            try:
                db.engine.execute(table.delete())
            except:
                log.debug("NOTE: couldn't delete this table")

            i = 0
            for row in reader:
                records.append(cls.make_record(row))
                i += 1
                if i >= batch_size:
                    db.engine.execute(table.insert(), records)
                    sys.stdout.write('*')
                    records = []
                    i = 0
            if len(records) > 0:
                db.engine.execute(table.insert(), records)
            f.close()

        # step 4: done...
        process_time = time.time() - start_time
        log.debug('{0}.load ({1:.0f} seconds)'.format(cls.__name__, process_time))

    @classmethod
    def post_process(cls, db, **kwargs):
        """
        Post-process processing method.  This method is a placeholder
        that may be overridden in children...
        @see: stop_time.py or route.py
        """
        pass

    @classmethod
    def make_record(cls, row):
        for k, v in row.copy().items():
            if isinstance(v, basestring):
                row[k] = v.strip()

            try:
                if k:
                    if (k not in cls.__table__.c):
                        del row[k]
                    elif not v:
                        row[k] = None
                    elif k.endswith('date'):
                        row[k] = datetime.datetime.strptime(v, '%Y%m%d').date()
                else:
                    log.info("I've got issues with your GTFS {0} data.  I'll continue, but expect more errors...".format(cls.__name__))
            except Exception as e:
                log.warning(e)

        """if this is a geospatially enabled database, add a geom"""
        if hasattr(cls, 'geom') and hasattr(cls, 'add_geom_to_dict'):
            cls.add_geom_to_dict(row)
        return row


Base = declarative_base(cls=_Base)



import sys
import time

from sqlalchemy import Column, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
from sqlalchemy.types import Date, Integer, String

from gtfsdb import config, util
from gtfsdb.model.base import Base
from .route_stop_base import RouteStopBase

import logging
log = logging.getLogger(__name__)


class RouteStop(Base, RouteStopBase):
    datasource = config.DATASOURCE_DERIVED
    __tablename__ = 'route_stops'

    id = Column(Integer, Sequence(None, optional=True), primary_key=True)
    route_id = Column(String(255), index=True, nullable=False)
    direction_id = Column(Integer, index=True, nullable=False)
    stop_id = Column(String(255), index=True, nullable=False)
    order = Column(Integer, index=True, nullable=False)
    start_date = Column(Date, index=True, nullable=False)
    end_date = Column(Date, index=True, nullable=False)

    route = relationship(
        'Route',
        primaryjoin='RouteStop.route_id==Route.route_id',
        foreign_keys='(RouteStop.route_id)',
        uselist=False, viewonly=True, lazy='joined')

    stop = relationship(
        'Stop',
        primaryjoin='RouteStop.stop_id==Stop.stop_id',
        foreign_keys='(RouteStop.stop_id)',
        uselist=False, viewonly=True, lazy='joined')

    direction = relationship(
        'RouteDirection',
        primaryjoin='RouteStop.route_id==RouteDirection.route_id and RouteStop.direction_id==RouteDirection.direction_id',
        foreign_keys='(RouteStop.route_id, RouteStop.direction_id)',
        uselist=False, viewonly=True, lazy='joined')

    start_calendar = relationship(
        'UniversalCalendar',
        primaryjoin='RouteStop.start_date==UniversalCalendar.date',
        foreign_keys='(RouteStop.start_date)',
        uselist=True, viewonly=True)

    end_calendar = relationship(
        'UniversalCalendar',
        primaryjoin='RouteStop.end_date==UniversalCalendar.date',
        foreign_keys='(RouteStop.end_date)',
        uselist=True, viewonly=True)

    def is_active(self, date=None):
        """
        :return False whenever we see that the route_stop's start and end date are
                outside the input date (where the input date defaults to 'today')
        """
        _is_active = False
        if self.start_date and self.end_date:
            date = util.check_date(date)
            if self.start_date <= date <= self.end_date:
                _is_active = True
        return _is_active

    def is_valid(self):
        ret_val = True
        if self.start_date is None or self.end_date is None:
            ret_val = False
        return ret_val

    def get_id(self):
        ret_val = "r:{0} d:{1} s:{2}".format(self.route_id, self.direction_id, self.stop_id)
        return ret_val

    @classmethod
    def is_stop_active(cls, session, stop_id, agency_id=None, date=None):
        """
        returns boolean whether given stop id is active for a given date
        """
        ret_val = False
        rs = RouteStop.query_by_stop(session, stop_id, agency_id, date, 1)
        if rs and len(rs) > 0:
            ret_val = True
        return ret_val

    @classmethod
    def active_unique_routes_at_stop(cls, session, stop_id, agency_id=None, date=None, route_name_filter=False):
        """
        to filter active routes, just provide a date to the above unique_routes_at_stop method
        """
        ret_val = []
        routes = cls.unique_routes_at_stop(session, stop_id, agency_id, date, route_name_filter)
        for r in routes:
            if r.is_active(date):
                ret_val.append(r)
        return ret_val

    @classmethod
    def query_active_stops(cls, session, route_id, direction_id=None, agency_id=None, date=None):
        """
        returns list of routes that are seen as 'active' based on dates and filters
        """
        # import pdb; pdb.set_trace()

        # step 1: default date
        date = util.check_date(date)

        # step 2a: query all route stops by route (and maybe direction and agency
        q = session.query(RouteStop).filter(RouteStop.route_id == route_id)
        if direction_id is not None:
            q = q.filter(RouteStop.direction_id == direction_id)
        if agency_id is not None:
            pass
            # TODO ... agency_id not in RouteStop -- should this even be here?
            # q = q.filter(RouteStop.agency_id == agency_id)

        # step 2b: filter based on date
        q = q.filter(RouteStop.start_date <= date).filter(date <= RouteStop.end_date)

        # step 2c: add some stop order
        q = q.order_by(RouteStop.order)

        route_stops = q.all()
        return route_stops

    @classmethod
    def load(cls, db, **kwargs):
        log.debug('{0}.load (loaded later in post_process)'.format(cls.__name__))
        pass

    @classmethod
    def post_process(cls, db, **kwargs):
        log.debug('{0}.post_process'.format(cls.__name__))
        cls.populate(db.session)

    @classmethod
    def populate(cls, session):
        """
        for each route/direction, find list of stop_ids for route/direction pairs

        the load is a two part process, where part A finds a list of unique stop ids, and
        part B creates the RouteStop (and potentially RouteDirections ... if not in GTFS) records
        """
        from gtfsdb import Route, RouteDirection

        start_time = time.time()
        routes = session.query(Route).all()

        for r in routes:
            # step 0: figure out some info about the route
            create_directions = False
            if r.directions is None or len(r.directions) == 0:
                create_directions = True

            # step 1a: filter the list of trips down to only a trip with a unique pattern
            trips = []
            shape_id_filter = []
            for t in r.trips:
                # a bit of a speedup to filter trips that have the same shape
                if t.shape_id and t.shape_id in shape_id_filter:
                    continue
                # store our trips
                shape_id_filter.append(t.shape_id)
                trips.append(t)

            # step 1b: sort our list of trips by length (note: for trips with two directions, ...)
            trips = sorted(trips, key=lambda t: t.trip_len, reverse=True)

            # step 2: get a hash table of route stops with effective start and end dates
            stop_effective_dates = cls._find_route_stop_effective_dates(session, r.route_id)

            # PART A: we're going to just collect a list of unique stop ids for this route / directions
            for d in [0, 1]:
                unique_stops = []

                # step 3: loop through all our trips and their stop times, pulling out a unique set of stops
                for t in trips:
                    if t.direction_id == d:

                        # step 4: loop through this trip's stop times, and find any/all stops that are in our stop list already
                        #         further, let's try to find the best position of that stop (e.g., look for where the stop patterns breaks)
                        last_pos = None
                        for i, st in enumerate(t.stop_times):
                            # step 5a: make sure this stop that customers can actually board...
                            if st.is_boarding_stop():

                                # step 5b: don't want arrival trips to influence route stop list
                                if st.stop_id in unique_stops:
                                    last_pos = unique_stops.index(st.stop_id)
                                else:
                                    # step 5b: add ths stop id to our unique list ... either in position, or appended to the end of the list
                                    if last_pos:
                                        last_pos += 1
                                        unique_stops.insert(last_pos, st.stop_id)
                                    else:
                                        unique_stops.append(st.stop_id)

                # PART B: add records to the database ...
                if len(unique_stops) > 0:

                    # step 6: if an entry for the direction doesn't exist, create a new
                    #         RouteDirection record and add it to this route
                    if create_directions:
                        rd = RouteDirection()
                        rd.route_id = r.route_id
                        rd.direction_id = d
                        rd.direction_name = "Outbound" if d == 0 else "Inbound"
                        session.add(rd)

                    # step 7: create new RouteStop records
                    for k, stop_id in enumerate(unique_stops):
                        # step 7: create a RouteStop record
                        rs = RouteStop()
                        rs.route_id = r.route_id
                        rs.direction_id = d
                        rs.stop_id = stop_id
                        rs.order = k + 1
                        s, e = cls._get_stop_effective_dates(stop_effective_dates, stop_id)
                        rs.start_date = s
                        rs.end_date = e
                        if rs.is_valid():
                            session.add(rs)
                        else:
                            log.info("{0} is not valid ... not adding to the database".format(rs.get_id()))

            # step 8: commit the new records to the db for this route...
            sys.stdout.write('*')
            session.commit()

        # step 9: final commit for any stragglers
        session.commit()
        session.flush()
        session.close()

        processing_time = time.time() - start_time
        log.debug('{0}.post_process ({1:.0f} seconds)'.format(cls.__name__, processing_time))

    @classmethod
    def _find_route_stop_effective_dates(cls, session, route_id):
        """
        find effective start date and end date for all stops of the input route, when
        queried against the trip and stop time tables.  Below are a couple of pure SQL queries that
        perform what I'm doing to get said start and end dates:

        # query all route stops with start & end dates
        SELECT t.route_id, st.stop_id, min(date), max(date)
        FROM ott.universal_calendar u, ott.trips t, ott.stop_times st
        WHERE t.service_id = u.service_id
          AND t.trip_id    = st.trip_id
        GROUP BY t.route_id, st.stop_id

        # query all stops start & end dates for a given route (used below in SQLAlchemy)
        SELECT st.stop_id, min(date), max(date)
        FROM ott.universal_calendar u, ott.trips t, ott.stop_times st
        WHERE t.service_id = u.service_id
          AND t.trip_id    = st.trip_id
          AND st.stop_id   = '1'
        GROUP BY st.stop_id

        :return hash table with stop_id as key, and tuple of (stop_id, start_date, end_date) for all route stops
        """
        ret_val = {}

        # step 1: query the route/stop start and end dates, based on stop time table
        from gtfsdb import UniversalCalendar, StopTime, Trip
        q = session.query(StopTime.stop_id, func.min(UniversalCalendar.date), func.max(UniversalCalendar.date))
        q = q.filter(UniversalCalendar.service_id == Trip.service_id)
        q = q.filter(Trip.trip_id  == StopTime.trip_id)
        q = q.filter(Trip.route_id == route_id)
        q = q.group_by(StopTime.stop_id)
        stop_dates = q.all()

        # step 2: make a hash of these dates with the stop id as the key
        for d in stop_dates:
            ret_val[d[0]] = d

        return ret_val

    @classmethod
    def _get_stop_effective_dates(cls, effective_dates_list, stop_id):
        """
        :return: start & end date from the route stop dates returned by method above
        :see: _find_route_stop_effective_dates
        """
        start = None
        end = None
        try:
            start = effective_dates_list[stop_id][1]
            end = effective_dates_list[stop_id][2]
        except Exception as e:
            log.info(e)
        return start, end


class CurrentRouteStops(Base, RouteStopBase):
    """
    this table is (optionally) used as a view into the currently active routes
    it is pre-calculated to list routes that are currently running service
    (GTFS can have multiple instances of the same route, with different aspects like name and direction)
    """
    datasource = config.DATASOURCE_DERIVED
    __tablename__ = 'current_route_stops'

    id = Column(Integer, primary_key=True, index=True, nullable=False)
    rs = relationship(
        'RouteStop',
        primaryjoin='CurrentRouteStops.id==RouteStop.id',
        foreign_keys='(RouteStop.id)',
        uselist=False, viewonly=True, lazy='joined'
    )

    route_id = Column(String(255), index=True, nullable=False)
    route = relationship(
        'Route',
        primaryjoin='Route.route_id==CurrentRouteStops.route_id',
        foreign_keys='(Route.route_id)',
        uselist=False, viewonly=True, lazy='joined'
    )

    stop_id = Column(String(255), index=True, nullable=False)
    stop = relationship(
        'Stop',
        primaryjoin='Stop.stop_id==CurrentRouteStops.stop_id',
        foreign_keys='(Stop.stop_id)',
        uselist=False, viewonly=True, lazy='joined'
    )

    order = Column(Integer, index=True, nullable=False)

    def __init__(self, route_stop):
        self.id = route_stop.id
        self.route_id = route_stop.route_id
        self.stop_id = route_stop.stop_id
        self.order = route_stop.order

    @classmethod
    def query_by_stop(cls, session, stop_id, agency_id=None, date=None, count=None, sort=False):
        """
        get all route stop records by looking for a given stop_id.
        further filtering can be had by providing an active date and agency id
        """
        # step 1: query stop id
        q = session.query(CurrentRouteStops).filter(CurrentRouteStops.stop_id == stop_id)
        if agency_id is not None:
            pass
            # TODO ... agency_id not in RouteStop -- should this even be here?
            # q = q.filter(RouteStop.agency_id == agency_id)

        # step 2: sort the results based on order column
        if sort:
            q = q.order_by(CurrentRouteStops.order)

        # step 3: limit the number of objects returned by query
        if count:
            q = q.limit(count)

        ret_val = q.all()
        return ret_val

    @classmethod
    def post_process(cls, db, **kwargs):
        """
        will update the current 'view' of this data
        """
        session = db.session()
        try:
            session.query(CurrentRouteStops).delete()

            rs_list = session.query(RouteStop).all()
            for rs in rs_list:
                if rs.is_active():
                    c = CurrentRouteStops(rs)
                    session.add(c)

            session.commit()
            session.flush()
        except Exception as e:
            log.warning(e)
            session.rollback()
        finally:
            session.flush()
            session.close()


__all__ = [RouteStop.__name__, CurrentRouteStops.__name__]



from gtfsdb import util
from gtfsdb.model.db import Database
from gtfsdb.api import database_load

from pkg_resources import resource_filename
import os
import logging
log = logging.getLogger(__name__)


def get_test_directory_path():
    """ will return current path ... tries to handle c:\\ windows junk """
    path = resource_filename('gtfsdb', 'tests')
    path = path.replace('c:\\', '/').replace('\\', '/')
    return path


def get_gtfs_file_uri(gtfs_file):
    """ will send back proper file:////blah/test_file.zip """
    dir_path = get_test_directory_path()
    file_uri = "file://{0}".format(os.path.join(dir_path, gtfs_file))
    file_uri = file_uri.replace('\\', '/')
    return file_uri


def load_sqlite(db_name=None, gtfs_name='multi-date-feed.zip'):
    # import pdb; pdb.set_trace()
    gtfs_uri = get_gtfs_file_uri(gtfs_name)
    url = util.make_temp_sqlite_db_uri(db_name)
    db = database_load(gtfs_uri, url=url, current_tables=True)
    return db


def load_pgsql(url, schema="current_test"):
    """ To run this test, do the following:
     x) bin/test  gtfsdb.tests.test_current

     You might also have to do the following:
     a) emacs setup.py - uncomment install_requires='psycopg2'
     b) buildout  # need psychopg2 in bin/test script
     c) comment out "#SKIP_TESTS = True" below
     d) psql -d postgres -c "CREATE DATABASE test WITH OWNER ott;"
     e) bin/test gtfsdb.tests.test_current
    """
    #import pdb; pdb.set_trace()
    gtfs_uri = get_gtfs_file_uri('multi-date-feed.zip')
    db = database_load(gtfs_uri, url=url, schema=schema, is_geospatial=True, current_tables=True)
    return db


def get_pg_db(url, schema='trimet'):
    db = Database(url=url, schema=schema, is_geospatial=True, current_tables=True)
    return db


def print_list(list):
    for i in list:
        print(i.__dict__)


def check_counts(list1, list2, id='stop_id'):
    """ check first that lists both have content; then chekc that either the lists are diff in size or content """
    ret_val = False
    #print_list(list1)
    #print_list(list2)
    if len(list1) > 0 and len(list2) > 0:
        if len(list1) != len(list2):
            ret_val = True
        else:
            for i, e1 in enumerate(list1):
                v1 = getattr(e1, id)
                v2 = getattr(list2[i], id)
                if v1 != v2:
                    ret_val = True
                    #print("{} VS. {}".format(v1, v2))
                    #print("{} VS. {}".format(e1.stop.stop_name, list2[i].stop_name))
                    break
    return ret_val
